# -*- coding: utf-8 -*-
"""
audit manager module.
"""

import traceback
import platform

import pyrin.application.services as application_services
import pyrin.packaging.services as packaging_services
import pyrin.globalization.datetime.services as datetime_services
import pyrin.configuration.services as config_services
import pyrin.globalization.locale.services as locale_services

from pyrin.audit.enumerations import InspectionStatusEnum
from pyrin.audit.hooks import AuditHookBase
from pyrin.core.mixin import HookMixin
from pyrin.core.structs import Manager
from pyrin.audit import AuditPackage
from pyrin.core.enumerations import ServerErrorResponseCodeEnum, SuccessfulResponseCodeEnum
from pyrin.audit.exceptions import InvalidAuditHookTypeError, AuditFailedError


class AuditManager(Manager, HookMixin):
    """
    audit manager class.

    this class provides some hooks to let different packages expose some audit services.
    it could be helpful to check system stability and status after a new deployment.
    """

    package_class = AuditPackage
    hook_type = AuditHookBase
    invalid_hook_type_error = InvalidAuditHookTypeError

    def _inspect_packages(self, **options):
        """
        this method will call `inspect` method of all registered hooks.

        it includes available info of all registered hooks, but it could be customized.
        each registered hook could be excluded from result if the related name of the hook
        is passed with False value. for example `database=False`.

        it returns a tuple of two values. first value is a dict containing the data provided
        by all registered hooks, and the second value is a bool value indicating that all
        hooks have been succeeded or not.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :keyword bool raise_error: specifies that it must raise error
                                   if any of registered audits failed
                                   instead of returning a failure response.
                                   defaults to False if not provided.

        :raises AuditFailedError: audit failed error.

        :rtype: tuple[dict, bool]
        """

        raise_error = options.get('raise_error', False)
        data = {}
        all_succeeded = True
        include_traceback = options.get('traceback', True)
        for hook in self._get_hooks():
            try:
                should_inspect = options.get(hook.audit_name, True)
                if should_inspect is True:
                    result, succeeded = hook.inspect(**options)
                    if succeeded is False:
                        all_succeeded = False
                    data[hook.audit_name] = result
            except Exception as error:
                if raise_error is True:
                    message = '[{name}] audit error: [{error}]'.format(name=hook.audit_name,
                                                                       error=str(error))
                    raise AuditFailedError(message) from error

                all_succeeded = False
                error_data = dict(status=InspectionStatusEnum.FAILED,
                                  error=str(error))
                if include_traceback is not False:
                    error_data.update(traceback=traceback.format_exc())
                data[hook.audit_name] = error_data

        return data, all_succeeded

    def inspect(self, **options):
        """
        inspects all registered packages and gets inspection data.

        it includes all available info, but it could be customized using different options.
        other custom implementations could also be excluded from result if the related name
        of their hook is passed with False value. for example `database=False`.

        it returns a tuple of two values. the first value is a dict containing the
        inspection data, and the second value is the related status code for the response.
        in case of any inspection has been failed, the status code will be internal
        server error 500.

        :keyword bool application: specifies that application info must be included.
        :keyword bool packages: specifies that loaded packages info must be included.
        :keyword bool framework: specifies that framework info must be included.
        :keyword bool python: specifies that python info must be included.
        :keyword bool os: specifies that operating system info must be included.
        :keyword bool hardware: specifies that hardware info must be included.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :keyword bool raise_error: specifies that it must raise error
                                   if any of registered audits failed
                                   instead of returning a failure response.
                                   defaults to False if not provided.

        :raises AuditFailedError: audit failed error.

        :returns: tuple[dict(dict application: application info,
                             dict packages: loaded packages info,
                             dict framework: framework info,
                             dict python: python info,
                             dict platform: platform info),
                        int status_code]
        :rtype: tuple[dict, int]
        """

        raise_error = options.get('raise_error', False)
        application = options.get('application', True)
        packages = options.get('packages', True)
        framework = options.get('framework', True)
        python = options.get('python', True)

        data, succeeded = self._inspect_packages(**options)

        if application is True:
            data.update(application=self.get_application_info(**options))

        if packages is True:
            packages = packaging_services.get_loaded_packages()
            packages_info = dict(names=packages,
                                 count=len(packages))
            data.update(packages=packages_info)

        if framework is True:
            data.update(framework=self.get_framework_info(**options))

        if python is True:
            data.update(python=self.get_python_info(**options))

        platform_info = self.get_platform_info(**options)
        if len(platform_info) > 0:
            data.update(platform=platform_info)

        if succeeded is False:
            if raise_error is False:
                return data, ServerErrorResponseCodeEnum.INTERNAL_SERVER_ERROR

            raise AuditFailedError('Audit has been failed.')

        return data, SuccessfulResponseCodeEnum.OK

    def startup_inspect(self):
        """
        inspects all registered packages on application startup.

        it raises an error if anything goes wrong.

        :raises AuditFailedError: audit failed error.
        """

        disabled = self.get_audit_configurations().get('ignore_startup') or []
        options = dict()
        for item in disabled:
            options[item] = False

        options.update(raise_error=True)
        self.inspect(**options)

    def get_application_info(self, **options):
        """
        gets the info of current application.

        :returns: dict(str name: application name,
                       datetime datetime: application current datetime,
                       str version: application version)
        :rtype: dict
        """

        data = {}
        data.update(name=application_services.get_application_name(),
                    version=application_services.get_application_version(),
                    server_datetime=datetime_services.now(),
                    server_timezone=datetime_services.get_timezone_name(server=True),
                    default_client_timezone=datetime_services.get_default_client_timezone().zone,
                    current_request_timezone=datetime_services.get_timezone_name(server=False),
                    default_locale=locale_services.get_default_locale(),
                    current_request_locale=locale_services.get_current_locale())
        return data

    def get_framework_info(self, **options):
        """
        gets the info of current framework.

        :returns: dict(str name: framework name,
                       str version: framework version)
        :rtype: dict
        """

        data = {}
        data.update(name='pyrin',
                    version=application_services.get_pyrin_version())
        return data

    def get_python_info(self, **options):
        """
        gets the current python version info which application is running on it.

        :returns: dict(str version: python version,
                       str implementation: python implementation)
        :rtype: dict
        """

        data = {}
        data.update(version=platform.python_version(),
                    implementation=platform.python_implementation())

        return data

    def get_operating_system_info(self, **options):
        """
        gets the current operating system info.

        :returns: dict(str name: os name,
                       str release: os release,
                       str version: os version)
        :rtype: dict
        """

        data = {}
        data.update(name=platform.system(),
                    release=platform.release(),
                    version=platform.version())

        return data

    def get_hardware_info(self, **options):
        """
        gets the current machine's hardware info.

        :returns: dict(str processor: processor name,
                       str machine: machine name)
        :rtype: dict
        """

        data = {}
        data.update(processor=platform.processor(),
                    machine=platform.machine())

        return data

    def get_platform_info(self, **options):
        """
        gets the platform info of current machine.

        it includes all available info, but it could be customized using different options.

        :keyword bool os: specifies that operating system info must be included.
        :keyword bool hardware: specifies that hardware info must be included.

        :returns: dict(dict os: os info,
                       dict hardware: hardware info)
        :rtype: dict
        """

        os = options.get('os', True)
        hardware = options.get('hardware', True)

        data = {}

        if os is True:
            data.update(os=self.get_operating_system_info(**options))

        if hardware is True:
            data.update(hardware=self.get_hardware_info(**options))

        return data

    def get_audit_configurations(self):
        """
        gets the audit api configurations.

        :returns: dict(bool enabled: enable audit api,
                       bool authenticated: audit api access type,
                       int request_limit: audit api request count limit,
                       str url: audit api exposed url)
        :rtype: dict
        """

        return config_services.get_active_section('audit')



# -*- coding: utf-8 -*-
"""
database session factory base module.
"""

from abc import abstractmethod

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.database.interface import AbstractSessionFactoryBase


class SessionFactoryBase(AbstractSessionFactoryBase):
    """
    session factory base class.
    """

    def __init__(self, **options):
        """
        initializes an instance of SessionFactoryBase.
        """

        super().__init__()

    def create_session_factory(self, engine):
        """
        creates a database session factory and binds it to given engine.

        :param Engine engine: database engine.

        :returns: database session
        :rtype: Session
        """

        session = self._create_session_factory(engine)
        setattr(session, 'session_factory_name', self.get_name())
        setattr(session, 'request_bounded', self.request_bounded)
        return session

    @abstractmethod
    def _create_session_factory(self, engine):
        """
        creates a database session factory and binds it to given engine.

        :param Engine engine: database engine.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: database session
        :rtype: Session
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
sequence component module.
"""

from pyrin.application.decorators import component
from pyrin.database.sequence import SequencePackage
from pyrin.database.sequence.manager import SequenceManager
from pyrin.application.structs import Component


@component(SequencePackage.COMPONENT_NAME)
class SequenceComponent(Component, SequenceManager):
    """
    sequence component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
configuration exceptions module.
"""

from pyrin.core.exceptions import CoreException


class ConfigurationManagerException(CoreException):
    """
    configuration manager exception.
    """
    pass


class ConfigurationStoreException(ConfigurationManagerException):
    """
    configuration store exception.
    """
    pass


class ConfigurationStoreAttributeNotFoundException(ConfigurationStoreException):
    """
    configuration store attribute not found exception.
    """
    pass


class ConfigurationStoreNotFoundError(ConfigurationManagerException):
    """
    configuration store not found error.
    """
    pass


class ConfigurationStoreExistedError(ConfigurationManagerException):
    """
    configuration store existed error.
    """
    pass


class ConfigurationFileNotFoundError(ConfigurationManagerException):
    """
    configuration file not found error.
    """
    pass


class ConfigurationStoreKeyNotFoundError(ConfigurationStoreAttributeNotFoundException):
    """
    configuration store key not found error.
    """
    pass


class ConfigurationStoreSectionNotFoundError(ConfigurationStoreAttributeNotFoundException):
    """
    configuration store section not found error.
    """
    pass


class ConfigurationStoreDuplicateKeyError(ConfigurationStoreException):
    """
    configuration store duplicate key error.
    """
    pass


class ConfigurationEnvironmentVariableNotFoundError(ConfigurationStoreException):
    """
    configuration environment variable not found error.
    """
    pass


class InvalidConfigurationEnvironmentVariableValueError(ConfigurationStoreException):
    """
    configuration environment variable value error.
    """
    pass


class ConfigurationFileExistedError(ConfigurationManagerException):
    """
    configuration file existed error.
    """
    pass


class InvalidPlaceHolderForEnvironmentVariableError(ConfigurationStoreException):
    """
    invalid place holder for environment variable error.
    """
    pass


class InvalidEnvironmentVariableKeyError(ConfigurationStoreException):
    """
    invalid environment variable key error.
    """
    pass



# -*- coding: utf-8 -*-
"""
alembic handlers show module.
"""

from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.enumerations import AlembicCLIHandlersEnum
from pyrin.database.migration.alembic.handlers.params import RevisionParam
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase


@alembic_cli_handler()
class ShowCLIHandler(AlembicCLIHandlerBase):
    """
    show cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of ShowCLIHandler.
        """

        super().__init__(AlembicCLIHandlersEnum.SHOW)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.append(RevisionParam())
        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
orm scoping interface module.
"""

from abc import abstractmethod

from pyrin.core.structs import CoreObject
from pyrin.core.exceptions import CoreNotImplementedError


class AbstractScopedRegistryBase(CoreObject):
    """
    abstract scoped registry base class.

    all application scoped registry classes must be subclassed from this.
    it also supports atomic objects.
    """

    @abstractmethod
    def __call__(self, atomic=False):
        """
        gets the corresponding object from registry if available, otherwise creates a new one.

        note that if there is an atomic object available, and `atomic=False` is
        provided, it always returns the available atomic object, but if `atomic=True`
        is provided it always creates a new atomic object.

        :param bool atomic: specifies that it must get a new atomic object.
                            defaults to False if not provided.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def has(self, atomic=False):
        """
        gets a value indicating that an object is present in the current scope.

        :param bool atomic: specifies that it must check just for an atomic
                            object. defaults to False if not provided.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def set(self, obj, atomic=False):
        """
        sets the value for the current scope.

        :param object obj: object to be set in registry.

        :param bool atomic: specifies that it must set an atomic object.
                            defaults to False if not provided.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def clear(self, atomic=False):
        """
        clears the current scope's objects, if any.

        it also clears if any atomic object is available.
        if `atomic=True` is provided, it only clears the
        current atomic object if available.

        :param bool atomic: specifies that it must only clear the current atomic
                            object of current scope. otherwise, it clears all objects
                            of current scope. defaults to False if not provided.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def clear_all_atomic(self):
        """
        clears all atomic objects of current scope, if any.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get(self, atomic=False):
        """
        gets the current object of current scope if available, otherwise returns None.

        :param bool atomic: specifies that it must get the current atomic object of
                            current scope, otherwise it returns the normal object.
                            defaults to False if not provided.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_all_atomic(self):
        """
        gets all atomic objects of current scope if available, otherwise returns an empty list.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: list[object]
        """

        raise CoreNotImplementedError()

    def inject_atomic(self, value, atomic=False):
        """
        injects the given atomic flag into given value and returns the same value.

        :param object value: value to inject atomic flag to it.

        :param bool atomic: indicates that the value should be marked
                            as atomic. defaults to False if not provided.

        :returns: object
        """

        setattr(value, 'atomic', atomic)
        return value



# -*- coding: utf-8 -*-
"""
encryption manager module.
"""

import pyrin.configuration.services as config_services

from pyrin.security.encryption import EncryptionPackage
from pyrin.security.encryption.interface import AbstractEncrypterBase
from pyrin.utils.custom_print import print_warning
from pyrin.core.structs import Context, Manager
from pyrin.security.encryption.exceptions import InvalidEncryptionHandlerTypeError, \
    InvalidEncryptionHandlerNameError, DuplicatedEncryptionHandlerError, \
    EncryptionHandlerNotFoundError, InvalidEncryptedValueError


class EncryptionManager(Manager):
    """
    encryption manager class.
    """

    package_class = EncryptionPackage

    def __init__(self):
        """
        initializes an instance of EncryptionManager.
        """

        super().__init__()

        self._encryption_handlers = Context()
        self._separator = '$'

    def register_encryption_handler(self, instance, **options):
        """
        registers a new encryption handler or replaces the existing one
        if `replace=True` is provided. otherwise, it raises an error
        on adding an instance which it's name is already available
        in registered handlers.

        :param AbstractEncrypterBase instance: encryption handler to be registered.
                                               it must be an instance of
                                               AbstractEncrypterBase.

        :keyword bool replace: specifies that if there is another registered
                               handler with the same name, replace it with
                               the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidEncryptionHandlerTypeError: invalid encryption handler type error.
        :raises InvalidEncryptionHandlerNameError: invalid encryption handler name error.
        :raises DuplicatedEncryptionHandlerError: duplicated encryption handler error.
        """

        if not isinstance(instance, AbstractEncrypterBase):
            raise InvalidEncryptionHandlerTypeError('Input parameter [{instance}] is '
                                                    'not an instance of [{base}].'
                                                    .format(instance=instance,
                                                            base=AbstractEncrypterBase))

        if instance.get_name() is None or len(instance.get_name().strip()) == 0:
            raise InvalidEncryptionHandlerNameError('Encryption handler [{instance}] '
                                                    'does not have a valid name.'
                                                    .format(instance=instance))

        # checking whether is there any registered instance with the same name.
        if instance.get_name() in self._encryption_handlers:
            replace = options.get('replace', False)

            if replace is not True:
                raise DuplicatedEncryptionHandlerError('There is another registered encryption '
                                                       'handler with name [{name}] but "replace" '
                                                       'option is not set, so handler '
                                                       '[{instance}] could not be registered.'
                                                       .format(name=instance.get_name(),
                                                               instance=instance))

            old_instance = self._encryption_handlers[instance.get_name()]
            print_warning('Encryption handler [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance, new_instance=instance))

        # registering new encryption handler.
        self._encryption_handlers[instance.get_name()] = instance

    def encrypt(self, text, **options):
        """
        encrypts the given value using specified handler and returns the encrypted result.

        :param str text: text to be encrypted.

        :keyword str handler_name: handler name to be used for encryption.
                                   if not provided, default handler from
                                   relevant configs will be used.

        :raises EncryptionHandlerNotFoundError: encryption handler not found error.

        :rtype: str
        """

        return self._get_encryption_handler(**options).encrypt(text, **options)

    def decrypt(self, full_encrypted_value, **options):
        """
        decrypts the given full encrypted value using specified
        handler and returns the decrypted result.

        :param str full_encrypted_value: full encrypted value to be decrypted.

        :raises InvalidEncryptionValueError: invalid encryption value error.
        :raises EncryptionHandlerNotFoundError: encryption handler not found error.
        :raises DecryptionError: decryption error.

        :rtype: str
        """

        handler_name = self._extract_handler_name(full_encrypted_value, **options)
        return self._get_encryption_handler(handler_name=handler_name).decrypt(
            full_encrypted_value, **options)

    def generate_key(self, handler_name, **options):
        """
        generates a valid key for the given handler and returns it.

        :param str handler_name: handler name to be used for key generation.

        :keyword int length: the length of generated key in bytes.
                             note that some encryption handlers may not accept custom
                             key length so this value would be ignored on those handlers.

        :raises EncryptionHandlerNotFoundError: encryption handler not found error.

        :rtype: str | tuple[str, str]
        """

        return self._get_encryption_handler(handler_name=handler_name).generate_key(**options)

    def _get_encryption_handler(self, **options):
        """
        gets the specified encryption handler.

        :keyword str handler_name: handler name to get.
                                   if not provided, default handler from
                                   relevant configs will be used.

        :raises EncryptionHandlerNotFoundError: encryption handler not found error.

        :rtype: AbstractEncrypterBase
        """

        handler_name = options.get('handler_name', self._get_default_handler_name())
        if handler_name not in self._encryption_handlers.keys():
            raise EncryptionHandlerNotFoundError('Encryption handler [{name}] not found.'
                                                 .format(name=handler_name))

        return self._encryption_handlers[handler_name]

    def _get_default_handler_name(self):
        """
        gets default encryption handler name from configs.

        :rtype: str
        """

        return config_services.get('security', 'encryption', 'default_encryption_handler')

    def _extract_handler_name(self, full_encrypted_value, **options):
        """
        extracts the handler name of given full encrypted value.

        :param str full_encrypted_value: full encrypted value to extract
                                         the handler name from.

        :raises InvalidEncryptionValueError: invalid encryption value error.

        :rtype: str
        """

        if full_encrypted_value.count(self._separator) < 2:
            raise InvalidEncryptedValueError('Input encrypted value has an incorrect format.')

        items = full_encrypted_value.split(self._separator, 2)
        return items[1]



# -*- coding: utf-8 -*-
"""
token services module.
"""

from pyrin.application.services import get_component
from pyrin.security.token import TokenPackage


def register_token_handler(instance, **options):
    """
    registers a new token handler or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding an instance which it's name is already available
    in registered handlers.

    :param AbstractTokenBase instance: token handler to be registered.
                                       it must be an instance of AbstractTokenBase.

    :keyword bool replace: specifies that if there is another registered
                           handler with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidTokenHandlerTypeError: invalid token handler type error.
    :raises InvalidTokenHandlerNameError: invalid token handler name error.
    :raises DuplicatedTokenHandlerError: duplicated token handler error.
    :raises DuplicatedTokenKidHeaderError: duplicated token kid header error.
    """

    return get_component(TokenPackage.COMPONENT_NAME).register_token_handler(instance,
                                                                             **options)


def generate_access_token(payload, **options):
    """
    generates an access token using specified handler from the given inputs and returns it.

    the generated token is in the form of `header_hash.payload_hash.signature_hash`
    and each part is encoded using a signing key.

    :param dict payload: a dictionary containing key/values as payload.
                         note that for better performance, keep payload
                         as small as possible.

    :keyword str handler_name: name of token handler to be used.
                               if not provided, default handler
                               from relevant configs will be used.

    :keyword dict custom_headers: a dictionary containing custom headers.

    :keyword bool is_fresh: indicates that this token is fresh.
                            being fresh means that token is generated by
                            providing user credentials to server.
                            if not provided, defaults to False.

    :raises TokenHandlerNotFoundError: token handler not found error.

    :returns: token.
    :rtype: str
    """

    return get_component(TokenPackage.COMPONENT_NAME).generate_access_token(payload, **options)


def generate_refresh_token(payload, **options):
    """
    generates a refresh token using specified handler from the given inputs and returns it.

    the generated token is in the form of `header_hash.payload_hash.signature_hash`
    and each part is encoded using a signing key.

    :param dict payload: a dictionary containing key/values as payload.
                         note that for better performance, keep payload
                         as small as possible.

    :keyword str handler_name: name of token handler to be used.
                               if not provided, default handler
                               from relevant configs will be used.

    :keyword dict custom_headers: a dictionary containing custom headers.

    :raises TokenHandlerNotFoundError: token handler not found error.

    :returns: token.
    :rtype: str
    """

    return get_component(TokenPackage.COMPONENT_NAME).generate_refresh_token(payload, **options)


def get_payload(token, **options):
    """
    decodes token using correct handler and gets the payload data.

    :param str token: token to get it's payload.

    :raises TokenKidHeaderNotSpecifiedError: token kid header not specified error.
    :raises TokenKidHeaderNotFoundError: token kid header not found error.
    :raises TokenHandlerNotFoundError: token handler not found error.
    :raises TokenVerificationError: token verification error.

    :rtype: dict
    """

    return get_component(TokenPackage.COMPONENT_NAME).get_payload(token, **options)


def get_unverified_payload(token, **options):
    """
    decodes token and gets the payload data without verifying the signature.

    note that returned payload must not be trusted for any critical operations.

    :param str token: token to get it's payload.

    :raises TokenKidHeaderNotSpecifiedError: token kid header not specified error.
    :raises TokenKidHeaderNotFoundError: token kid header not found error.
    :raises TokenHandlerNotFoundError: token handler not found error.
    :raises TokenDecodingError: token decoding error.

    :rtype: dict
    """

    return get_component(TokenPackage.COMPONENT_NAME).get_unverified_payload(token, **options)


def get_unverified_header(token, **options):
    """
    gets the header dict of token without verifying the signature.

    note that the returned header must not be trusted for critical operations.

    :param str token: token to get it's header.

    :raises TokenDecodingError: token decoding error.

    :rtype: dict
    """

    return get_component(TokenPackage.COMPONENT_NAME).get_unverified_header(token, **options)


def generate_key(handler_name, **options):
    """
    generates a valid key for the given handler and returns it.

    :param str handler_name: token handler name to be used.

    :keyword int length: the length of generated key in bytes.
                         note that some token handlers may not accept custom
                         key length so this value would be ignored on those handlers.

    :rtype: str | tuple[str, str]
    """

    return get_component(TokenPackage.COMPONENT_NAME).generate_key(handler_name, **options)



# -*- coding: utf-8 -*-
"""
database thread_scoped session factory module.
"""

from sqlalchemy.orm import sessionmaker

import pyrin.configuration.services as config_services

from pyrin.database.decorators import session_factory
from pyrin.database.orm.query.base import CoreQuery
from pyrin.database.orm.scoping.base import CoreScopedSession
from pyrin.database.orm.session.base import CoreSession
from pyrin.database.session_factory.base import SessionFactoryBase


@session_factory()
class ThreadScopedSessionFactory(SessionFactoryBase):
    """
    thread scoped session factory class.
    """

    def __init__(self, **options):
        """
        initializes an instance of ThreadScopedSessionFactory.
        """

        super().__init__(**options)

    def _create_session_factory(self, engine):
        """
        creates a database thread scoped session factory and binds it to given engine.

        the scope is current thread.

        :param Engine engine: database engine.

        :returns: database session
        :rtype: Session
        """

        session_configs = config_services.get_section('database', 'thread_scoped_session')
        return CoreScopedSession(sessionmaker(bind=engine, class_=CoreSession,
                                              query_cls=CoreQuery, **session_configs))

    @property
    def request_bounded(self):
        """
        gets a value indicating that this session factory type should be bounded into request.

        :rtype: bool
        """

        return False



# -*- coding: utf-8 -*-
"""
hashing component module.
"""

from pyrin.application.decorators import component
from pyrin.security.hashing import HashingPackage
from pyrin.security.hashing.manager import HashingManager
from pyrin.application.structs import Component


@component(HashingPackage.COMPONENT_NAME)
class HashingComponent(Component, HashingManager):
    """
    hashing component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
model manager module.
"""

import pyrin.utils.sqlalchemy as sqlalchemy_utils

from pyrin.core.mixin import HookMixin
from pyrin.core.structs import Manager
from pyrin.database.model import ModelPackage
from pyrin.database.model.base import CoreEntity
from pyrin.database.model.hooks import ModelHookBase
from pyrin.database.model.exceptions import EntitiesAreNotCollectedError, \
    InvalidModelHookTypeError


class ModelManager(Manager, HookMixin):
    """
    model manager class.
    """

    hook_type = ModelHookBase
    invalid_hook_type_error = InvalidModelHookTypeError
    package_class = ModelPackage

    def __init__(self):
        """
        initializes an instance of ModelManager.
        """

        super().__init__()

        # a tuple containing all application entity classes.
        self._entities = None

    def _after_entities_collected(self):
        """
        this will call `after_entities_collected` method of registered hooks.
        """

        for hook in self._get_hooks():
            hook.after_entities_collected()

    def get_declarative_base(self):
        """
        gets the application's declarative base class.

        by default it returns `CoreEntity` class.
        if you want to define a new declarative base class in your application
        instead of using `CoreEntity` which is provided by pyrin, you must
        override this method inside your application's `database.model.ModelManager`
        class and return your new declarative base class.
        this is required if you want to enable migrations in your application.
        note that your application must have a unique declarative base type and
        you shouldn't mix the usage of `CoreEntity` and your new declarative base,
        otherwise migrations will not work properly.

        but, if you don't want to use migrations at all, you could just put
        `pyrin.database.migration` into `ignored_packages` list of `packaging.ini`
        file and leave this method unimplemented.

        :rtype: type[BaseEntity]
        """

        return CoreEntity

    def get_metadata(self):
        """
        gets metadata of current declarative base.

        :rtype: MetaData
        """

        return self.get_declarative_base().metadata

    def get_tables(self):
        """
        gets all tables defined in metadata of current declarative base.

        :returns: dict(str name, Table table)
        :rtype: dict
        """

        return self.get_metadata().tables

    def collect_entities(self):
        """
        collects all entity classes of application.

        it returns the count of collected entities.

        :rtype: int
        """

        base = self.get_declarative_base()
        result = []
        for table in self.get_tables().values():
            entity = sqlalchemy_utils.get_class_by_table(base, table, raise_multi=False)
            if entity is not None:
                result.append(entity)

        self._entities = tuple(result)
        self._after_entities_collected()
        return len(result)

    def get_entities(self):
        """
        gets a tuple of all application collected entities.

        :raises EntitiesAreNotCollectedError: entities are not collected error.

        :rtype: tuple[BaseEntity]
        """

        if self._entities is None:
            raise EntitiesAreNotCollectedError('Application entities are not collected yet.')

        return self._entities



# -*- coding: utf-8 -*-
"""
template services module.
"""

from pyrin.application.services import get_component
from pyrin.template import TemplatePackage


def register_template_handler(instance, **options):
    """
    registers a new template handler or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding a template handler which is already registered.

    :param AbstractTemplateHandler instance: template handler to be registered.
                                             it must be an instance of
                                             AbstractTemplateHandler.

    :keyword bool replace: specifies that if there is another registered
                           template handler with the same name, replace it
                           with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidTemplateHandlerTypeError: invalid template handler type error.
    :raises DuplicatedTemplateHandlerError: duplicated template handler error.
    """

    return get_component(TemplatePackage.COMPONENT_NAME).register_template_handler(instance,
                                                                                   **options)


def create(handler_name, *args, **kwargs):
    """
    creates the template using the given template handler name.

    :param str handler_name: handler name to be used.
    :param object args: arguments that should be passed to template handler.

    :keyword object kwargs: keyword arguments that should be passed to template handler.

    :raises TemplateHandlerNotFoundError: template handler not found error.
    """

    return get_component(TemplatePackage.COMPONENT_NAME).create(handler_name, *args, **kwargs)



# -*- coding: utf-8 -*-
"""
string normalizer handlers replace module.
"""

from pyrin.utilities.string.normalizer.decorators import string_normalizer
from pyrin.utilities.string.normalizer.enumerations import NormalizerEnum
from pyrin.utilities.string.normalizer.handlers.base import ReplaceNormalizerBase


@string_normalizer()
class PersianNumberNormalizer(ReplaceNormalizerBase):
    """
    persian number normalizer class.

    this normalizer replaces all persian numbers with latin numbers.
    """

    def __init__(self, **options):
        """
        initializes an instance of PersianNumberNormalizer.
        """

        replace_map = {'۰': '0', '۱': '1', '۲': '2', '۳': '3', '۴': '4',
                       '۵': '5', '۶': '6', '۷': '7', '۸': '8', '۹': '9'}

        super().__init__(NormalizerEnum.PERSIAN_NUMBER, 310, replace_map, **options)


@string_normalizer()
class ArabicNumberNormalizer(ReplaceNormalizerBase):
    """
    arabic number normalizer class.

    this normalizer replaces all arabic numbers with latin numbers.
    """

    def __init__(self, **options):
        """
        initializes an instance of ArabicNumberNormalizer.
        """

        replace_map = {'٠': '0', '١': '1', '٢': '2', '٣': '3', '٤': '4',
                       '٥': '5', '٦': '6', '٧': '7', '٨': '8', '٩': '9'}

        super().__init__(NormalizerEnum.ARABIC_NUMBER, 320, replace_map, **options)


@string_normalizer()
class PersianLetterNormalizer(ReplaceNormalizerBase):
    """
    persian letter normalizer class.

    this normalizer replaces all non standard persian letters with standard versions.
    """

    def __init__(self, **options):
        """
        initializes an instance of PersianLetterNormalizer.
        """

        replace_map = {'ي': 'ی', 'ك': 'ک', 'إ': 'ا', 'أ': 'ا',
                       'ؤ': 'و', 'ۀ': 'ه', 'ة': 'ه', 'آ': 'ا'}

        super().__init__(NormalizerEnum.PERSIAN_LETTER, 330, replace_map, **options)


@string_normalizer()
class LatinLetterNormalizer(ReplaceNormalizerBase):
    """
    latin letter normalizer class.

    this normalizer replaces all latin letters with english versions.
    """

    def __init__(self, **options):
        """
        initializes an instance of LatinLetterNormalizer.
        """

        replace_map = {'ä': 'a', 'Ä': 'A', 'ë': 'e', 'Ë': 'E', 'ï': 'i',
                       'Ï': 'I', 'ö': 'o', 'Ö': 'O', 'ü': 'u', 'Ü': 'U',
                       'ÿ': 'y', 'Ÿ': 'Y', 'á': 'a', 'Á': 'A', 'ć': 'c',
                       'Ć': 'C', 'é': 'e', 'É': 'E', 'í': 'i', 'Í': 'I',
                       'ń': 'n', 'Ń': 'N', 'ó': 'o', 'Ó': 'O', 'ś': 's',
                       'Ś': 'S', 'ú': 'u', 'Ú': 'U', 'ý': 'y', 'Ý': 'Y',
                       'ź': 'z', 'Ź': 'Z', 'ő': 'o', 'Ő': 'O', 'ű': 'u',
                       'Ű': 'U', 'à': 'a', 'À': 'A', 'è': 'e', 'È': 'E',
                       'ì': 'i', 'Ì': 'I', 'ò': 'o', 'Ò': 'O', 'ù': 'u',
                       'Ù': 'U', 'â': 'a', 'Â': 'A', 'ê': 'e', 'Ê': 'E',
                       'î': 'i', 'Î': 'I', 'ô': 'o', 'Ô': 'O', 'û': 'u',
                       'Û': 'U', 'ã': 'a', 'Ã': 'A', 'ñ': 'n', 'Ñ': 'N',
                       'õ': 'o', 'Õ': 'O', 'č': 'c', 'Č': 'C', 'ď': 'd',
                       'Ď': 'D', 'ě': 'e', 'Ě': 'E', 'ǧ': 'g', 'Ǧ': 'G',
                       'ň': 'n', 'Ň': 'N', 'ř': 'r', 'Ř': 'R', 'š': 's',
                       'Š': 'S', 'ť': 't', 'Ť': 'T', 'ž': 'z', 'Ž': 'Z',
                       'đ': 'd', 'Đ': 'D'}

        super().__init__(NormalizerEnum.LATIN_LETTER, 340, replace_map, **options)



# -*- coding: utf-8 -*-
"""
locale package.
"""

from pyrin.packaging.base import Package


class LocalePackage(Package):
    """
    locale package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'globalization.locale.component'
    DEPENDS = ['pyrin.configuration']



# -*- coding: utf-8 -*-
"""
caching local containers base module.
"""

from abc import abstractmethod

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.structs import CoreObject


class LocalCacheContainerBase(CoreObject):
    """
    local cache container base class.

    it has a similar interface to a normal dict except it has an extra `set` method.
    all application local cache containers must be subclassed from this.
    """

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """
        implements call on this class type.

        in the form of `LocalCacheContainerBase()`.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def __setitem__(self, key, value):
        """
        sets the given key with given value into the cache.

        :param object key: key to set value with it.
        :param object value: value to be cached.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def __getitem__(self, key):
        """
        gets the value of given key.

        in the form of `self[key]`.
        it raises an error if the key does not exist.

        :param object key: key to get its value.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def __delitem__(self, key):
        """
        deletes the given key from cache.

        in the form of `del self[key]`.
        it raises an error if the key does not exist.

        :param object key: key to be deleted.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def __iter__(self, *args, **kwargs):
        """
        iterates over current cached items.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: iterator
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def __contains__(self, key):
        """
        gets a value indicating that given key exists in cache.

        :param object key: key to be checked for existence.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def __len__(self):
        """
        gets the count of items of this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def set(self, key, value, *args, **kwargs):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get(self, key, default=None):
        """
        gets the value from cache.

        if key does not exist, it returns None or the specified default value.

        :param object key: hashable key to get its value from cache.
        :param object default: value to be returned if key is not present.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def contains(self, key):
        """
        gets a value indicating that given key is existed in the cached items.

        :param object key: key to be checked for existence.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def pop(self, key, default=None):
        """
        pops the given key from cache and returns its value.

        if key does not exist, it returns None or the specified default value.

        :param object key: key to get its value.
        :param object default: value to be returned if key is not present.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def remove(self, key):
        """
        removes the given key from cache.

        it does nothing if the key is not present in the cache.

        :param object key: key to be removed.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def clear(self):
        """
        clears all items from cache.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def items(self):
        """
        gets an iterable of all keys and their values in the cache.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: iterable[tuple[object, object]]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def keys(self):
        """
        gets all keys of current cache.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: iterable[object]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def values(self):
        """
        gets all values of current cache.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: iterable[object]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def move_to_end(self, key, last=True):
        """
        moves the given key and its value to the end of queue.

        :param object key: key to be moved.

        :param bool last: specifies that item must be moved to end.
                          defaults to True if not provided.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def slice(self, count, last=False):
        """
        gets a slice of items of this container with the count length.

        :param int count: number of items to include in slice.

        :param bool last: specifies that slice must include items from end.
                          defaults to False if not provided.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: iterable
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
deserializer handlers string module.
"""

import re

from pyrin.converters.deserializer.decorators import deserializer
from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase


@deserializer()
class StringDeserializer(StringPatternDeserializerBase):
    """
    string deserializer class.
    """

    # default min for this deserializer is 2 because it should
    # at least have two single or double quotes at both ends.
    DEFAULT_MIN = 2

    # matches a string value, all of these values will be matched.
    # example: "sample_value", "12345678", 'true'
    # note that the second and third examples will be deserialized as a
    # string with value 12345678 and true and won't deserialize as an
    # integer or boolean, because of the enclosing single or double quotes.
    DOUBLE_QUOTE_REGEX = re.compile(r'^[\"].*[\"]$')
    SINGLE_QUOTE_REGEX = re.compile(r'^[\'].*[\']$')

    def __init__(self, **options):
        """
        creates an instance of StringDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  patterns and their min and
                                                                  max length for string
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: str
        """

        # removing the first and last single or double quotes from value.
        value = value[1:-1]
        return value

    @property
    def default_formats(self):
        """
        gets default accepted patterns that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(self.DOUBLE_QUOTE_REGEX, self.UNDEF_LENGTH, self.UNDEF_LENGTH),
                (self.SINGLE_QUOTE_REGEX, self.UNDEF_LENGTH, self.UNDEF_LENGTH)]



# -*- coding: utf-8 -*-
"""
celery cli component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.task_queues.celery.cli import CeleryCLIPackage
from pyrin.task_queues.celery.cli.manager import CeleryCLIManager


@component(CeleryCLIPackage.COMPONENT_NAME)
class CeleryCLIComponent(Component, CeleryCLIManager):
    """
    celery cli component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
request enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class RequestHeaderEnum(CoreEnum):
    """
    request header enum.
    """

    # A-IM: feed
    # instance manipulations that are acceptable in the response.
    A_IM = 'A-IM'

    # Accept: application/json
    # the media type/types acceptable.
    ACCEPT = 'Accept'

    # Accept-Charset: utf-8
    # the charset acceptable.
    ACCEPT_CHARSET = 'Accept-Charset'

    # Accept-Encoding: gzip, deflate
    # list of acceptable encodings.
    ACCEPT_ENCODING = 'Accept-Encoding'

    # Accept-Language: en-US
    # list of acceptable languages.
    ACCEPT_LANGUAGE = 'Accept-Language'

    # Accept-Datetime: Thu, 31 May 2007 20:35:00 GMT
    # request a past version of the resource prior to the datetime passed.
    ACCEPT_DATETIME = 'Accept-Datetime'

    # Access-Control-Request-Method: GET
    # used in a CORS request.
    ACCESS_CONTROL_REQUEST_METHOD = 'Access-Control-Request-Method'

    # Access-Control-Request-Headers: origin, x-requested-with, accept
    # used in a CORS request.
    ACCESS_CONTROL_REQUEST_HEADERS = 'Access-Control-Request-Headers'

    # Authorization: Basic 34i3j4iom2323==
    # HTTP basic authentication credentials.
    AUTHORIZATION = 'Authorization'

    # Cache-Control: no-cache
    # set the caching rules.
    CACHE_CONTROL = 'Cache-Control'

    # Connection: keep-alive
    # control options for the current connection.
    # Accepts 'keep-alive' and 'close'.
    # deprecated in HTTP/2.
    CONNECTION = 'Connection'

    # Content-Length: 348
    # the length of the request body in bytes.
    CONTENT_LENGTH = 'Content-Length'

    # Content-Type: application/x-www-form-urlencoded
    # the content type of the body of the request (used in POST and PUT requests).
    CONTENT_TYPE = 'Content-Type'

    # Cookie: name=value; Secure; HttpOnly
    COOKIE = 'Cookie'

    # Date: Tue, 15 Nov 1994 08:12:31 GMT
    # the date and time that the request was sent.
    DATE = 'Date'

    # Expect: 100-continue
    # it’s typically used when sending a large request body. we expect the server to
    # return back a '100 Continue' HTTP status if it can handle the request, or
    # '417 Expectation Failed' if not.
    EXPECT = 'Expect'

    # Forwarded: for=192.0.2.60; proto=http; by=203.0.113.43
    # disclose original information of a client connecting to a web server through an
    # HTTP proxy. used for testing purposes only, as it discloses privacy sensitive information.
    FORWARDED = 'Forwarded'

    # From: user@example.com
    # the email address of the user making the request. meant to be used, for example,
    # to indicate a contact email for bots.
    FROM = 'From'

    # Host: flaviocopes.com
    # the domain name of the server (used to determined the server with virtual hosting), and
    # the TCP port number on which the server is listening. if the port is omitted, 80 is
    # assumed. this is a mandatory HTTP request header.
    HOST = 'Host'

    # If-Match: "737060cd8c284d8582d"
    # given one (or more) ETags, the server should only send back the response if the
    # current resource matches one of those ETags. mainly used in PUT methods to update
    # a resource only if it has not been modified since the user last updated it.
    IF_MATCH = 'If-Match'

    # If-Modified-Since: Sat, 29 Oct 1994 19:43:31 GMT
    # allows to return a '304 Not Modified' response header if the content is unchanged
    # since that date.
    IF_MODIFIED_SINCE = 'If-Modified-Since'

    # If-None-Match: "737060cd882f209582d"
    # allows a '304 Not Modified' response header to be returned if content is unchanged.
    # opposite of If-Match.
    IF_NONE_MATCH = 'If-None-Match'

    # If-Range: "737060cd8c9582d"
    # used to resume downloads, returns a partial if the condition is matched (ETag or date)
    # or the full resource if not.
    IF_RANGE = 'If-Range'

    # If-Unmodified-Since: Sat, 29 Oct 1994 19:43:31 GMT
    # only send the response if the entity has not been modified since the specified time.
    IF_UNMODIFIED_SINCE = 'If-Unmodified-Since'

    # Max-Forwards: 10
    # limit the number of times the message can be forwarded through proxies or gateways.
    MAX_FORWARDS = 'Max-Forwards'

    # Origin: http://mydomain.com
    # send the current domain to perform a CORS request, used in an OPTIONS HTTP request.
    # (to ask the server for Access-Control- response headers).
    ORIGIN = 'Origin'

    # Pragma: no-cache
    # used for backwards compatibility with HTTP/1.0 caches.
    PRAGMA = 'Pragma'

    # Proxy-Authorization: Basic 2323jiojioIJOIOJIJ==
    # authorization credentials for connecting to a proxy.
    PROXY_AUTHORIZATION = 'Proxy-Authorization'

    # Range: bytes=500-999
    # request only a specific part of a resource.
    RANGE = 'Range'

    # Referer: https://flaviocopes.com
    # the address of the previous web page from which a link to the currently requested
    # page was followed.
    REFERER = 'Referer'

    # TE: trailers, deflate
    # specify the encodings the client can accept.
    # accepted values: compress, deflate, gzip, trailers.
    # only trailers is supported in HTTP/2.
    TE = 'TE'

    # User-Agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6)
    # AppleWebKit/537.36 (KHTML, like Gecko) Chrome/67.0.3396.99 Safari/537.36
    # the string that identifies the user agent.
    USER_AGENT = 'User-Agent'

    # Upgrade: h2c, HTTPS/1.3, IRC/6.9, RTA/x11, websocket
    # ask the server to upgrade to another protocol.
    # deprecated in HTTP/2.
    UPGRADE = 'Upgrade'

    # Via: 1.0 fred, 1.1 example.com (Apache/1.1)
    # informs the server of proxies through which the request was sent.
    VIA = 'Via'

    # Warning: 199 Miscellaneous warning
    # a general warning about possible problems with the status of the message.
    # see https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Warning
    # for range of values.
    WARNING = 'Warning'


class NonStandardRequestHeaderEnum(CoreEnum):
    """
    non standard request header enum.
    """

    # DNT: 1
    # if enabled, asks servers to not track the user.
    DNT = 'Dnt'

    # X-Requested-With: XMLHttpRequest
    # identifies XHR requests.
    X_REQUESTED_WITH = 'X-Requested-With'

    # X-CSRF-Token: <TOKEN>
    # used to prevent CSRF.
    X_CSRF_TOKEN = 'X-CSRF-Token'



# -*- coding: utf-8 -*-
"""
caching local items complex module.
"""

from pyrin.caching.local.items.base import ComplexLocalCacheItemBase


class ComplexLocalCacheItem(ComplexLocalCacheItemBase):
    """
    complex local cache item class.

    this type of cache item supports expire time.
    it also keeps the deep copy of the value into the cache.
    """
    pass



# -*- coding: utf-8 -*-
"""
utils custom_print module.
"""

import colorama

import pyrin.application.services as application_services


def print_colorful(value, color, force=False):
    """
    prints the given value into stdout using the given color.
    if the application has been started in scripting mode
    it does not print the value.

    :param object value: value to be printed.

    :param int color: color of text to be printed.
                      it should be from `colorama.Fore` colors.

    :param bool force: forces the printing, even if application
                       has been started in scripting mode.
                       defaults to False if not provided.
    """

    if force is True or application_services.is_scripting_mode() is False:
        try:
            if not isinstance(value, str):
                value = str(value)
            colorama.init(autoreset=True)
            print(str(color) + value)
        finally:
            colorama.deinit()


def print_warning(value, force=False):
    """
    prints the given value into stdout as a warning.
    if the application has been started in scripting mode
    it does not print the value.

    :param object value: value to be printed.

    :param bool force: forces the printing, even if application
                       has been started in scripting mode.
                       defaults to False if not provided.
    """

    print_colorful(value, colorama.Fore.YELLOW, force)


def print_error(value, force=False):
    """
    prints the given value into stdout as an error.
    if the application has been started in scripting mode
    it does not print the value.

    :param object value: value to be printed.

    :param bool force: forces the printing, even if application
                       has been started in scripting mode.
                       defaults to False if not provided.
    """

    print_colorful(value, colorama.Fore.RED, force)


def print_info(value, force=False):
    """
    prints the given value into stdout as an info.
    if the application has been started in scripting mode
    it does not print the value.

    :param object value: value to be printed.

    :param bool force: forces the printing, even if application
                       has been started in scripting mode.
                       defaults to False if not provided.
    """

    print_colorful(value, colorama.Fore.BLUE, force)


def print_default(value, force=False):
    """
    prints the given value into stdout with default color.
    if the application has been started in scripting mode
    it does not print the value.

    :param object value: value to be printed.

    :param bool force: forces the printing, even if application
                       has been started in scripting mode.
                       defaults to False if not provided.
    """

    print_colorful(value, colorama.Fore.RESET, force)



# -*- coding: utf-8 -*-
"""
caching globals module.
"""

# this value must be passed to caches when
# no count limit is required.
NO_LIMIT = 'No Limit'

# these are common types that are not hashable and require conversion.
# note that despite tuple is hashable itself, but we have to validate all
# of its internal values to be also hashable. so we have to put tuple in this list.
NOT_HASHABLE_TYPES = (list, dict, set, tuple)



# -*- coding: utf-8 -*-
"""
database migration services module.
"""

from pyrin.application.services import get_component
from pyrin.database.migration import DatabaseMigrationPackage


def create_all():
    """
    creates all entities on database engine.
    """

    return get_component(DatabaseMigrationPackage.COMPONENT_NAME).create_all()


def drop_all():
    """
    drops all entities on database engine.
    """

    return get_component(DatabaseMigrationPackage.COMPONENT_NAME).drop_all()


def get_connection_urls():
    """
    gets all database connection urls for each engine.

    it gets the values from active section of each store.
    in case of sqlite usage, all urls will be made absolute.

    :returns: dict[str bind_name: str connection_url]
    :rtype: dict
    """

    return get_component(DatabaseMigrationPackage.COMPONENT_NAME).get_connection_urls()


def get_bind_name_to_metadata_map():
    """
    gets bind name to metadata map.

    :returns: dict[str bind_name: MetaData metadata]
    :rtype: dict
    """

    return get_component(DatabaseMigrationPackage.COMPONENT_NAME).get_bind_name_to_metadata_map()


def configure_migration_data():
    """
    configures the required data for any migration.
    """

    return get_component(DatabaseMigrationPackage.COMPONENT_NAME).configure_migration_data()


def rebuild_schema():
    """
    rebuilds database schema based on database and environment config store values.

    this will occur only if application is not in scripting mode.
    """

    return get_component(DatabaseMigrationPackage.COMPONENT_NAME).rebuild_schema()



# -*- coding: utf-8 -*-
"""
validator interface module.
"""

from abc import abstractmethod

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.structs import CoreObject


class AbstractValidatorBase(CoreObject):
    """
    abstract validator base class.

    all application validators must be subclassed from this.
    """

    @abstractmethod
    def validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        it returns the same or fixed value.

        :param object | list[object] value: value to be validated.

        :keyword bool for_update: specifies that this field is being
                                  validated for update operation.
                                  defaults to False if not provided.

        :raises CoreNotImplementedError: core not implemented error.
        :raises ValidationError: validation error.

        :returns: object | list[object]
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def name(self):
        """
        gets the name of this validator.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def domain(self):
        """
        gets the domain of this validator.

        domain is the type of a BaseEntity subclass that
        this validator validates a value of it.
        if a validator is not specific to an entity, then
        domain could be a unique string name.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type[BaseEntity] | str
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
serializer package.
"""

from pyrin.packaging.base import Package


class SerializerPackage(Package):
    """
    serializer package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'converters.serializer.component'



# -*- coding: utf-8 -*-
"""
cli exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class CLIManagerException(CoreException):
    """
    cli manager exception.
    """
    pass


class CLIManagerBusinessException(CoreBusinessException, CLIManagerException):
    """
    cli manager business exception.
    """
    pass


class ArgumentParamNameIsRequiredError(CLIManagerException):
    """
    argument param name is required error.
    """
    pass


class PositionalArgumentIndexError(CLIManagerException):
    """
    positional argument index error.
    """
    pass


class BooleanArgumentValueError(CLIManagerException):
    """
    boolean argument value error.
    """
    pass


class KeywordArgumentCLIOptionNameError(CLIManagerException):
    """
    keyword argument cli option name error.
    """
    pass


class MappingArgumentParamValueToCLIMapRequiredError(CLIManagerException):
    """
    mapping argument param value to cli map required error.
    """
    pass


class KeywordArgumentCLIOptionNameRequiredError(CLIManagerException):
    """
    keyword argument cli option name required error.
    """
    pass


class InvalidCLIHandlerNameError(CLIManagerException):
    """
    invalid cli handler name error.
    """
    pass


class InvalidCLIHandlerTypeError(CLIManagerException):
    """
    invalid cli handler type error.
    """
    pass


class DuplicatedCLIHandlerError(CLIManagerException):
    """
    duplicated cli handler error.
    """
    pass


class CLIHandlerNotFoundError(CLIManagerBusinessException):
    """
    cli handler not found error.
    """
    pass


class PositionalArgumentsIndicesError(CLIManagerException):
    """
    positional arguments indices error.
    """
    pass


class InvalidCLIDecoratedMethodError(CLIManagerException):
    """
    invalid cli decorated method error.
    """
    pass


class InvalidCLIParamTypeError(CLIManagerException):
    """
    invalid cli param type error.
    """
    pass


class DuplicatedCLIGroupError(CLIManagerException):
    """
    duplicated cli group error.
    """
    pass


class InvalidCLIGroupTypeError(CLIManagerException):
    """
    invalid cli group type error.
    """
    pass


class CLIGroupNameIsRequiredError(CLIManagerException):
    """
    cli group name is required error.
    """
    pass



# -*- coding: utf-8 -*-
"""
validator handlers misc module.
"""

from pyrin.core.globals import _, LIST_TYPES
from pyrin.validator.handlers.base import ValidatorBase
from pyrin.validator.handlers.exceptions import ValueIsLowerThanMinimumError, \
    ValueIsHigherThanMaximumError, ValueIsOutOfRangeError, \
    AcceptedMinimumValueMustBeProvidedError, AcceptedMaximumValueMustBeProvidedError, \
    ValidValuesMustBeProvidedError, InvalidValuesMustBeProvidedError, \
    MinimumValueLargerThanMaximumValueError


class MinimumValidator(ValidatorBase):
    """
    minimum validator class.
    """

    minimum_value_error = ValueIsLowerThanMinimumError
    minimum_value_message = _('The provided value for [{param_name}] must '
                              'be greater than {or_equal}[{minimum}].')

    inclusive_minimum_value_message = _('or equal to ')

    # accepted minimum value, it could also be a callable without any inputs.
    default_accepted_minimum = None
    default_inclusive_minimum = None

    def __init__(self, domain, field, **options):
        """
        initializes an instance of MinimumValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param InstrumentedAttribute | str field: validator field name. it could be a
                                                  string or a column. each validator will
                                                  be registered with its field name in
                                                  corresponding domain. to enable automatic
                                                  validations, the provided field name must
                                                  be the exact name of the parameter which
                                                  this validator will validate. if you pass
                                                  a column attribute, some constraints
                                                  such as `nullable`, `min_length`, `max_length`,
                                                  `min_value`, `max_value`, `allow_blank`,
                                                  `allow_whitespace`, `check_in` and
                                                  `check_not_in` could be extracted
                                                  automatically from that column if not provided
                                                  in inputs.

        :keyword type | tuple[type] accepted_type: accepted type for value.
                                                   no type checking will be
                                                   done if not provided.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this value has precedence over `inclusive_minimum`
                                         instance attribute if provided.

        :raises ValidatorFieldIsRequiredError: validator field is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidAcceptedTypeError: invalid accepted type error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        :raises AcceptedMinimumValueMustBeProvidedError: accepted minimum value
                                                         must be provided error.
        """

        super().__init__(domain, field, **options)

        inclusive_minimum = options.get('inclusive_minimum')
        if inclusive_minimum is None:
            if self.default_inclusive_minimum is not None:
                inclusive_minimum = self.default_inclusive_minimum
            else:
                inclusive_minimum = True

        if self.default_accepted_minimum is None and \
                (self.field is None or self.field.min_value is None):
            raise AcceptedMinimumValueMustBeProvidedError('Accepted minimum value '
                                                          'could not be None.')

        if self.default_accepted_minimum is None:
            self.default_accepted_minimum = self.field.min_value
        else:
            self.default_accepted_minimum = self.default_accepted_minimum

        self._inclusive_minimum = inclusive_minimum

        self._validate_exception_type(self.minimum_value_error)

    def _validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate()`
        preferably at the beginning.

        :param object value: value to be validated.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this value has precedence over `inclusive_minimum`
                                         instance attribute if provided.

        :raises ValueIsLowerThanMinimumError: value is lower than minimum error.
        """

        super()._validate(value, **options)

        inclusive_minimum = options.get('inclusive_minimum')
        if inclusive_minimum is None:
            inclusive_minimum = self.inclusive_minimum

        current_min = self.accepted_minimum
        current_values = options.get(RangeValidator.CURRENT_VALUE_KEY)
        if current_values is not None:
            current_values[RangeValidator.CURRENT_MIN_KEY] = current_min

        if value < current_min or (value == current_min and inclusive_minimum is False):
            equality = ''
            if inclusive_minimum is not False:
                equality = self.inclusive_minimum_value_message

            raise self.minimum_value_error(
                self.minimum_value_message.format(param_name=self.localized_name,
                                                  minimum=
                                                  self._get_representation(current_min),
                                                  or_equal=equality))

    @property
    def accepted_minimum(self):
        """
        gets the lower bound of values that this validator considers valid.

        :rtype: object
        """

        return self._get_value(self.default_accepted_minimum)

    @property
    def inclusive_minimum(self):
        """
        gets a value indicating that values equal to accepted minimum must be considered valid.

        :rtype: bool
        """

        return self._inclusive_minimum


class MaximumValidator(ValidatorBase):
    """
    maximum validator class.
    """

    maximum_value_error = ValueIsHigherThanMaximumError
    maximum_value_message = _('The provided value for [{param_name}] must '
                              'be lower than {or_equal}[{maximum}].')

    inclusive_maximum_value_message = _('or equal to ')

    # accepted maximum value, it could also be a callable without any inputs.
    default_accepted_maximum = None
    default_inclusive_maximum = None

    def __init__(self, domain, field, **options):
        """
        initializes an instance of MaximumValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param InstrumentedAttribute | str field: validator field name. it could be a
                                                  string or a column. each validator will
                                                  be registered with its field name in
                                                  corresponding domain. to enable automatic
                                                  validations, the provided field name must
                                                  be the exact name of the parameter which
                                                  this validator will validate. if you pass
                                                  a column attribute, some constraints
                                                  such as `nullable`, `min_length`, `max_length`,
                                                  `min_value`, `max_value`, `allow_blank`,
                                                  `allow_whitespace`, `check_in` and
                                                  `check_not_in` could be extracted
                                                  automatically from that column if not provided
                                                  in inputs.

        :keyword type | tuple[type] accepted_type: accepted type for value.
                                                   no type checking will be
                                                   done if not provided.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this value has precedence over `inclusive_maximum`
                                         instance attribute if provided.

        :raises ValidatorFieldIsRequiredError: validator field is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidAcceptedTypeError: invalid accepted type error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        :raises AcceptedMaximumValueMustBeProvidedError: accepted maximum value
                                                         must be provided error.
        """

        super().__init__(domain, field, **options)

        inclusive_maximum = options.get('inclusive_maximum')
        if inclusive_maximum is None:
            if self.default_inclusive_maximum is not None:
                inclusive_maximum = self.default_inclusive_maximum
            else:
                inclusive_maximum = True

        if self.default_accepted_maximum is None and \
                (self.field is None or self.field.max_value is None):
            raise AcceptedMaximumValueMustBeProvidedError('Accepted maximum value '
                                                          'could not be None.')

        if self.default_accepted_maximum is None:
            self.default_accepted_maximum = self.field.max_value
        else:
            self.default_accepted_maximum = self.default_accepted_maximum

        self._inclusive_maximum = inclusive_maximum

        self._validate_exception_type(self.maximum_value_error)

    def _validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate()`
        preferably at the beginning.

        :param object value: value to be validated.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this value has precedence over `inclusive_maximum`
                                         instance attribute if provided.

        :raises ValueIsHigherThanMaximumError: value is higher than maximum error.
        """

        super()._validate(value, **options)

        inclusive_maximum = options.get('inclusive_maximum')
        if inclusive_maximum is None:
            inclusive_maximum = self.inclusive_maximum

        current_max = self.accepted_maximum
        current_values = options.get(RangeValidator.CURRENT_VALUE_KEY)
        if current_values is not None:
            current_values[RangeValidator.CURRENT_MAX_KEY] = current_max

        if value > current_max or (value == current_max and inclusive_maximum is False):
            equality = ''
            if inclusive_maximum is not False:
                equality = self.inclusive_maximum_value_message

            raise self.maximum_value_error(
                self.maximum_value_message.format(param_name=self.localized_name,
                                                  maximum=
                                                  self._get_representation(current_max),
                                                  or_equal=equality))

    @property
    def accepted_maximum(self):
        """
        gets the upper bound of values that this validator considers valid.

        :rtype: object
        """

        return self._get_value(self.default_accepted_maximum)

    @property
    def inclusive_maximum(self):
        """
        gets a value indicating that values equal to accepted maximum must be considered valid.

        :rtype: bool
        """

        return self._inclusive_maximum


class RangeValidator(MinimumValidator, MaximumValidator):
    """
    range validator class.
    """

    # these values are used inside minimum, maximum and range validators.
    CURRENT_VALUE_KEY = '__current__'
    CURRENT_MAX_KEY = 'max'
    CURRENT_MIN_KEY = 'min'

    range_value_error = ValueIsOutOfRangeError
    range_value_message = _('The provided value for [{param_name}] must '
                            'be greater than {or_equal_min}[{lower}] and '
                            'lower than {or_equal_max}[{upper}].')

    def __init__(self, domain, field, **options):
        """
        initializes an instance of RangeValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param InstrumentedAttribute | str field: validator field name. it could be a
                                                  string or a column. each validator will
                                                  be registered with its field name in
                                                  corresponding domain. to enable automatic
                                                  validations, the provided field name must
                                                  be the exact name of the parameter which
                                                  this validator will validate. if you pass
                                                  a column attribute, some constraints
                                                  such as `nullable`, `min_length`, `max_length`,
                                                  `min_value`, `max_value`, `allow_blank`,
                                                  `allow_whitespace`, `check_in` and
                                                  `check_not_in` could be extracted
                                                  automatically from that column if not provided
                                                  in inputs.

        :keyword type | tuple[type] accepted_type: accepted type for value.
                                                   no type checking will be
                                                   done if not provided.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this value has precedence over `inclusive_minimum`
                                         instance attribute if provided.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this value has precedence over `inclusive_maximum`
                                         instance attribute if provided.

        :raises ValidatorFieldIsRequiredError: validator field is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidAcceptedTypeError: invalid accepted type error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        :raises AcceptedMinimumValueMustBeProvidedError: accepted minimum value
                                                         must be provided error.
        :raises AcceptedMaximumValueMustBeProvidedError: accepted maximum value
                                                         must be provided error.
        :raises MinimumValueLargerThanMaximumValueError: minimum value larger
                                                         than maximum value error.
        """

        super().__init__(domain, field, **options)

        if not callable(self.default_accepted_minimum) and \
                not callable(self.default_accepted_maximum) and \
                self.accepted_minimum is not None and self.accepted_maximum is not None \
                and self.accepted_minimum > self.accepted_maximum:
            raise MinimumValueLargerThanMaximumValueError('Accepted minimum value could not be '
                                                          'larger than accepted maximum value.')

        self._validate_exception_type(self.range_value_error)

    def _validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate()`
        preferably at the beginning.

        :param object value: value to be validated.

        :keyword bool inclusive_minimum: determines that values equal to
                                         accepted minimum should be considered valid.
                                         this value has precedence over `inclusive_minimum`
                                         instance attribute if provided.

        :keyword bool inclusive_maximum: determines that values equal to
                                         accepted maximum should be considered valid.
                                         this value has precedence over `inclusive_maximum`
                                         instance attribute if provided.

        :raises ValueIsOutOfRangeError: value is out of range error.
        """

        # this is a workaround to get the correct values of accepted min and max in
        # case they are callable and producing different results on each call.
        current_values = dict()
        current_values[self.CURRENT_MAX_KEY] = None
        current_values[self.CURRENT_MIN_KEY] = None
        options[self.CURRENT_VALUE_KEY] = current_values
        try:
            super()._validate(value, **options)
        except (self.maximum_value_error, self.minimum_value_error):
            equality_min = ''
            equality_max = ''

            inclusive_maximum = options.get('inclusive_maximum')
            if inclusive_maximum is None:
                inclusive_maximum = self.inclusive_maximum

            inclusive_minimum = options.get('inclusive_minimum')
            if inclusive_minimum is None:
                inclusive_minimum = self.inclusive_minimum

            if inclusive_minimum is not False:
                equality_min = self.inclusive_minimum_value_message

            if inclusive_maximum is not False:
                equality_max = self.inclusive_maximum_value_message

            current_min = current_values.get(self.CURRENT_MIN_KEY)
            if current_min is None:
                current_min = self.accepted_minimum

            current_max = current_values.get(self.CURRENT_MAX_KEY)
            if current_max is None:
                current_max = self.accepted_maximum

            raise self.range_value_error(self.range_value_message.format(
                param_name=self.localized_name,
                lower=self._get_representation(current_min),
                upper=self._get_representation(current_max),
                or_equal_min=equality_min, or_equal_max=equality_max))


class InValidator(ValidatorBase):
    """
    in validator class.
    """

    not_in_value_error = ValueIsOutOfRangeError
    not_in_value_message = _('The provided value for [{param_name}] '
                             'must be one of {values}.')

    # valid values, it must be an iterable or a callable.
    default_valid_values = None

    def __init__(self, domain, field, **options):
        """
        initializes an instance of InValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param InstrumentedAttribute | str field: validator field name. it could be a
                                                  string or a column. each validator will
                                                  be registered with its field name in
                                                  corresponding domain. to enable automatic
                                                  validations, the provided field name must
                                                  be the exact name of the parameter which
                                                  this validator will validate. if you pass
                                                  a column attribute, some constraints
                                                  such as `nullable`, `min_length`, `max_length`,
                                                  `min_value`, `max_value`, `allow_blank`,
                                                  `allow_whitespace`, `check_in` and
                                                  `check_not_in` could be extracted
                                                  automatically from that column if not provided
                                                  in inputs.

        :keyword type | tuple[type] accepted_type: accepted type for value.
                                                   no type checking will be
                                                   done if not provided.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :raises ValidatorFieldIsRequiredError: validator field is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidAcceptedTypeError: invalid accepted type error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        :raises ValidValuesMustBeProvidedError: valid values must be provided error.
        """

        super().__init__(domain, field, **options)

        if self.default_valid_values is None \
                and self.field is not None and self.field.check_in is not None:
            self.default_valid_values = self.field.check_in

        if not callable(self.default_valid_values) and \
                (self.default_valid_values is None or
                 not isinstance(self.default_valid_values, LIST_TYPES) or
                 len(self.default_valid_values) <= 0):
            raise ValidValuesMustBeProvidedError('Valid values must be provided '
                                                 'as iterable or callable.')

        self._validate_exception_type(self.not_in_value_error)

    def _validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate()`
        preferably at the beginning.

        :param object value: value to be validated.

        :raises ValueIsOutOfRangeError: value is out of range error.
        """

        super()._validate(value, **options)

        current_valid = self.valid_values
        if value not in current_valid:
            raise self.not_in_value_error(self.not_in_value_message.format(
                param_name=self.localized_name,
                values=self._get_list_representation(current_valid)))

    @property
    def valid_values(self):
        """
        gets a list of valid values for this validator.

        :rtype: list[object]
        """

        return self._get_value(self.default_valid_values)


class NotInValidator(ValidatorBase):
    """
    not in validator class.
    """

    in_value_error = ValueIsOutOfRangeError
    in_value_message = _('The provided value for [{param_name}] '
                         'could not be one of {values}.')

    # invalid values, it must be an iterable or a callable.
    default_invalid_values = None

    def __init__(self, domain, field, **options):
        """
        initializes an instance of NotInValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param InstrumentedAttribute | str field: validator field name. it could be a
                                                  string or a column. each validator will
                                                  be registered with its field name in
                                                  corresponding domain. to enable automatic
                                                  validations, the provided field name must
                                                  be the exact name of the parameter which
                                                  this validator will validate. if you pass
                                                  a column attribute, some constraints
                                                  such as `nullable`, `min_length`, `max_length`,
                                                  `min_value`, `max_value`, `allow_blank`,
                                                  `allow_whitespace`, `check_in` and
                                                  `check_not_in` could be extracted
                                                  automatically from that column if not provided
                                                  in inputs.

        :keyword type | tuple[type] accepted_type: accepted type for value.
                                                   no type checking will be
                                                   done if not provided.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :raises ValidatorFieldIsRequiredError: validator field is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidAcceptedTypeError: invalid accepted type error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        :raises InvalidValuesMustBeProvidedError: invalid values must be provided error.
        """

        super().__init__(domain, field, **options)

        if self.default_invalid_values is None \
                and self.field is not None and self.field.check_not_in is not None:
            self.default_invalid_values = self.field.check_not_in

        if not callable(self.default_invalid_values) and \
                (self.default_invalid_values is None or
                 not isinstance(self.default_invalid_values, LIST_TYPES) or
                 len(self.default_invalid_values) <= 0):
            raise InvalidValuesMustBeProvidedError('Invalid values must be '
                                                   'provided as iterable or callable.')

        self._validate_exception_type(self.in_value_error)

    def _validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate()`
        preferably at the beginning.

        :param object value: value to be validated.

        :raises ValueIsOutOfRangeError: value is out of range error.
        """

        super()._validate(value, **options)

        current_invalid = self.invalid_values
        if value in current_invalid:
            raise self.in_value_error(self.in_value_message.format(
                param_name=self.localized_name,
                values=self._get_list_representation(current_invalid)))

    @property
    def invalid_values(self):
        """
        gets a list of invalid values for this validator.

        :rtype: list[object]
        """

        return self._get_value(self.default_invalid_values)



# -*- coding: utf-8 -*-
"""
celery decorators module.
"""

import pyrin.task_queues.celery.services as celery_services


# you could use this as decorator to register your tasks.
# you could also use celery's '@shared_task' decorator if you prefer.
task = celery_services.get_current_app().task



# -*- coding: utf-8 -*-
"""
logging adapters module.
"""

from logging import Logger, raiseExceptions

import pyrin.security.session.services as session_services
import pyrin.logging.services as logging_services
import pyrin.utils.string as string_utils

from pyrin.logging.enumerations import LogLevelIntEnum


class BaseLoggerAdapter(Logger):
    """
    base logger adapter class.

    all logger adapters should be subclassed from this.

    this class is intended to be an adapter for `Logger`, but it is implemented
    this way because of two different issues:

    1. if we implement it as an stand-alone adapter and not subclassing it from
    `Logger`, then logging package will raise an error on wrapped loggers where
    it checks the type of logger against `Logger`. for example in fixup methods.

    2. if we fully subclass this from `Logger` and removing the adapter behavior,
    another issue arises and in different situations the generated logs will not
    be emitted correctly. because the original logger with the same name exists
    before instantiating the adapter.

    so we have to implement this class this way, subclass it from `Logger`, but in
    fact use it as an adapter and proxying all accesses to `Logger` attributes to the
    underlying actual `Logger` instance.

    also note that we *DO NOT* call `super().__init()` when initializing the adapter.
    because the actual logger is the `_logger` attribute and all operations will be
    proxied to it.
    """

    def __init__(self, logger):
        """
        initializes an instance of BaseLoggerAdapter.

        :param Logger logger: logger instance to be wrapped.
        """

        self.extra = {}
        self._logger = logger

    def __str__(self):
        """
        gets string representation of this logger.

        :rtype: str
        """

        return str(self._logger)

    def __repr__(self):
        """
        gets string representation of this logger.

        :rtype: str
        """

        return repr(self._logger)

    def __reduce__(self):
        """
        gets the reduced version of this logger.

        :rtype: tuple
        """

        return self._logger.__reduce__()

    def log(self, level, msg, *args, **kwargs):
        """
        delegates a log call to the underlying logger.

        this method is overridden to execute before and after emit hooks.
        and also performing interpolation data preparation if required.

        :param int level: log level.
        :param str msg: log message.
        :param dict kwargs: keyword arguments passed to logging call.

        :keyword dict interpolation_data: data to be used for interpolation.

        :raises TypeError: type error.
        """

        if not isinstance(level, int):
            if raiseExceptions:
                raise TypeError('log level must be an integer.')
            else:
                return

        if self.isEnabledFor(level):
            self._log(level, msg, args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        """
        log message with `DEBUG` severity.

        to pass exception information, use the keyword argument `exc_info` with
        a true value.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        if self.isEnabledFor(LogLevelIntEnum.DEBUG):
            self._log(LogLevelIntEnum.DEBUG, msg, args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        log message with `INFO` severity.

        to pass exception information, use the keyword argument `exc_info` with
        a true value.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        if self.isEnabledFor(LogLevelIntEnum.INFO):
            self._log(LogLevelIntEnum.INFO, msg, args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        log message with `WARNING` severity.

        to pass exception information, use the keyword argument `exc_info` with
        a true value.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        if self.isEnabledFor(LogLevelIntEnum.WARNING):
            self._log(LogLevelIntEnum.WARNING, msg, args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        """
        log message with `WARNING` severity.

        to pass exception information, use the keyword argument `exc_info` with
        a true value.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        self.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        log message with `ERROR` severity.

        to pass exception information, use the keyword argument `exc_info` with
        a true value.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        if self.isEnabledFor(LogLevelIntEnum.ERROR):
            self._log(LogLevelIntEnum.ERROR, msg, args, **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs):
        """
        log exception with `ERROR` severity.

        :param str msg: log message.
        :param bool exc_info: include exception info in log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        self.error(msg, *args, exc_info=exc_info, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        log message with `CRITICAL` severity.

        to pass exception information, use the keyword argument `exc_info` with
        a true value.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        if self.isEnabledFor(LogLevelIntEnum.CRITICAL):
            self._log(LogLevelIntEnum.CRITICAL, msg, args, **kwargs)

    def fatal(self, msg, *args, **kwargs):
        """
        log message with `CRITICAL` severity.

        to pass exception information, use the keyword argument `exc_info` with
        a true value.

        :param str msg: log message.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        self.critical(msg, *args, **kwargs)

    def process(self, msg, kwargs):
        """
        processes the logging message and keyword arguments passed in to a logging call.

        it's to insert contextual information. you can either manipulate
        the message itself, the keyword args or both. return the message
        and kwargs modified (or not) to suit your needs.

        :param str msg: log message.
        :param dict kwargs: keyword arguments passed to logging call.

        :returns: tuple[str message, dict kwargs]
        :rtype: tuple[str, dict]
        """

        kwargs["extra"] = self.extra
        return self._process(msg, kwargs)

    def setLevel(self, level):
        """
        sets the logging level of this logger. level must be an int or a str.

        :param int | str level: logging level to be set.
        """

        self._logger.setLevel(level)

    def findCaller(self, stack_info=False, stacklevel=1):
        """
        finds the stack frame of the caller.

        so that we can note the source file name, line number and function name.

        :param bool stack_info: include stack info.
        :param int stacklevel: stack level to be included.

        :rtype: tuple
        """

        return self._logger.findCaller(stack_info, stacklevel)

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info,
                   func=None, extra=None, sinfo=None):
        """
        a factory method which can be overridden in subclasses to create
        specialized `LogRecords`.

        :rtype: LogRecord
        """

        return self._logger.makeRecord(name, level, fn, lno, msg, args, exc_info,
                                       func=func, extra=extra, sinfo=sinfo)

    def handle(self, record):
        """
        calls the handlers for the specified record.

        this method is used for unpickled records received from a socket, as
        well as those created locally. logger-level filtering is applied.

        :param LogRecord record: log record instance.
        """

        self._logger.handle(record)

    def addHandler(self, hdlr):
        """
        adds the specified handler to this logger.
        """

        self._logger.addHandler(hdlr)

    def removeHandler(self, hdlr):
        """
        removes the specified handler from this logger.
        """

        self._logger.removeHandler(hdlr)

    def hasHandlers(self):
        """
        gets a value indicating that this logger has any configured handlers.

        loop through all handlers for this logger and its parents in the
        logger hierarchy. return true if a handler was found, else false.
        stop searching up the hierarchy whenever a logger with the `propagate`
        attribute set to zero is found - that will be the last logger which
        is checked for the existence of handlers.

        :rtype: bool
        """

        return self._logger.hasHandlers()

    def callHandlers(self, record):
        """
        passes a record to all relevant handlers.

        loop through all handlers for this logger and its parents in the
        logger hierarchy. if no handler was found, output a one-off error
        message to sys.stderr. stop searching up the hierarchy whenever a
        logger with the 'propagate' attribute set to zero is found - that
        will be the last logger whose handlers are called.

        :param LogRecord record: log record instance.
        """

        self._logger.callHandlers(record)

    def getEffectiveLevel(self):
        """
        gets the effective level for this logger.

        loop through this logger and its parents in the logger hierarchy,
        looking for a non-zero logging level. return the first one found.

        :rtype: int
        """

        return self._logger.getEffectiveLevel()

    def isEnabledFor(self, level):
        """
        gets a value indicating that this logger is enabled for given log level.

        :param int level: logging level.

        :rtype: bool
        """

        return self._logger.isEnabledFor(level)

    def getChild(self, suffix):
        """
        gets a logger which is a descendant to this one.

        this is a convenience method, such that:

        `logging.getLogger('abc').getChild('def.ghi')`

        is the same as:

        `logging.getLogger('abc.def.ghi')`

        it's useful, for example, when the parent logger is named using
        `__name__` rather than a literal string.
        """

        return self._logger.getChild(suffix)

    def _log(self, level, msg, args, exc_info=None, extra=None,
             stack_info=False, stacklevel=1, **options):
        """
        low-level logging routine which creates a `LogRecord`.

        and then calls all the handlers of this logger to handle the record.

        :param int level: logging level.
        :param str msg: log message.
        :param tuple args: logging arguments.
        :param bool exc_info: include exception info in log message.
        :param dict extra: extra keyword arguments.
        :param bool stack_info: include stack info.
        :param int stacklevel: stack level to be included.

        :keyword dict interpolation_data: data to be used for interpolation.
        """

        data = options.pop('interpolation_data', None)
        if data is not None:
            data = logging_services.prepare_data(data)
            msg = string_utils.interpolate(msg, data)

        custom_message, custom_kwargs = self.process(msg, options)
        logging_services.before_emit(custom_message, data, level, **custom_kwargs)
        self._logger._log(level, custom_message, args, exc_info=exc_info,
                          extra=extra, stack_info=stack_info, stacklevel=stacklevel)
        logging_services.after_emit(custom_message, data, level, **custom_kwargs)

    def _process(self, msg, kwargs):
        """
        processes the logging message and keyword arguments passed in to a logging call.

        it's to insert contextual information. you can either manipulate
        the message itself, the keyword args or both. return the message
        and kwargs modified (or not) to suit your needs.

        this method is intended to be overridden in subclasses.

        :param str msg: log message.
        :param dict kwargs: keyword arguments passed to logging call.

        :returns: tuple[str message, dict kwargs]
        :rtype: tuple[str, dict]
        """

        return msg, kwargs

    @property
    def name(self):
        """
        gets the name of this logger.

        :rtype: str
        """

        return self._logger.name

    @name.setter
    def name(self, value):
        """
        sets the name of this logger.

        :param str value: value to be set as name.
        """

        self._logger.name = value

    @property
    def level(self):
        """
        gets the level of this logger.

        :rtype: int
        """

        return self._logger.level

    @level.setter
    def level(self, value):
        """
        sets the level of this logger.

        :param int | str value: value to be set as level.
        """

        self._logger.level = value

    @property
    def parent(self):
        """
        gets the parent logger of this logger.

        :rtype: Logger
        """

        return self._logger.parent

    @parent.setter
    def parent(self, value):
        """
        sets the parent logger of this logger.

        :param Logger value: value to be set as parent.
        """

        self._logger.parent = value

    @property
    def propagate(self):
        """
        gets a value indicating that this logger must propagate exceptions.

        :rtype: bool
        """

        return self._logger.propagate

    @propagate.setter
    def propagate(self, value):
        """
        sets the propagate value of this logger.

        :param bool value: value to be set as propagate.
        """

        self._logger.propagate = value

    @property
    def handlers(self):
        """
        gets the handlers of this logger.

        :rtype: list
        """

        return self._logger.handlers

    @handlers.setter
    def handlers(self, value):
        """
        sets the handlers of this logger.

        :param list value: values to be set as handlers.
        """

        self._logger.handlers = value

    @property
    def disabled(self):
        """
        gets a value indicating that this logger is disabled.

        :rtype: bool
        """

        return self._logger.disabled

    @disabled.setter
    def disabled(self, value):
        """
        sets the disabled value of this logger.

        :param bool value: value to be set as disabled.
        """

        self._logger.disabled = value

    @property
    def _cache(self):
        """
        gets the cache of this logger.

        :rtype: dict
        """

        return self._logger._cache

    @_cache.setter
    def _cache(self, value):
        """
        sets the cache of this logger.

        :param dict value: value to be set as cache.
        """

        self._logger._cache = value


class RequestInfoLoggerAdapter(BaseLoggerAdapter):
    """
    request info logger adapter.

    this adapter adds request info into generated logs.
    """

    def _process(self, msg, kwargs):
        """
        processes the logging message and keyword arguments passed in to a logging call.

        it's to insert contextual information. you can either manipulate
        the message itself, the keyword args or both. return the message
        and kwargs modified (or not) to suit your needs.

        :param str msg: log message.
        :param dict kwargs: keyword arguments passed to logging call.

        :returns: tuple[str message, dict kwargs]
        :rtype: tuple[str, dict]
        """

        client_request = session_services.get_safe_current_request()
        request_info = ''
        if client_request is not None:
            request_info = '[{client_request}]: '.format(client_request=client_request)

        custom_message = '{request_info}{message}'.format(request_info=request_info,
                                                          message=msg)
        return custom_message, kwargs



# -*- coding: utf-8 -*-
"""
mimetype exceptions module.
"""

from pyrin.core.exceptions import CoreException


class MIMETypeManagerException(CoreException):
    """
    mimetype manager exception.
    """
    pass


class InvalidMIMETypeHandlerTypeError(MIMETypeManagerException):
    """
    invalid mimetype handler type error.
    """
    pass


class DuplicatedMIMETypeHandlerError(MIMETypeManagerException):
    """
    duplicated mimetype handler error.
    """
    pass



# -*- coding: utf-8 -*-
"""
orm scoping registry request_scoped module.
"""

from sqlalchemy.util import ScopedRegistry

import pyrin.security.session.services as session_services

from pyrin.database.orm.scoping.interface import AbstractScopedRegistryBase
from pyrin.database.orm.scoping.registry.containers.atomic import RequestScopedAtomicContainer


class RequestScopedRegistry(ScopedRegistry, AbstractScopedRegistryBase):
    """
    request scoped registry class.

    this class provides a way to overcome the out of request
    context error when configuring sessions on application startup.
    it also supports atomic objects.
    """

    def __init__(self, createfunc, scopefunc):
        """
        initializes and instance of RequestScopedRegistry.

        :param function createfunc: a creation function that will generate a new
                                    value for the current scope, if none is present.

        :param function scopefunc: a function that returns a hashable
                                   token representing the current scope.
                                   such as, current thread identifier.
        """

        super().__init__(createfunc, scopefunc)

        self.atomic_registry = RequestScopedAtomicContainer(scopefunc)

    def __call__(self, atomic=False):
        """
        gets the corresponding object from registry if available, otherwise creates a new one.

        note that if there is an atomic object available, and `atomic=False` is
        provided, it always returns the available atomic object, but if `atomic=True`
        is provided it always creates a new atomic object.

        :param bool atomic: specifies that it must get a new atomic object.
                            defaults to False if not provided.

        :returns: object
        """

        if atomic is True:
            value = self.inject_atomic(self.createfunc(), atomic)
            self.set(value, atomic)
            return value

        value = self.atomic_registry.get()
        if value is not None:
            return value

        return self.inject_atomic(super().__call__(), atomic)

    def has(self, atomic=False):
        """
        gets a value indicating that an object is present in the current scope.

        :param bool atomic: specifies that it must check just for an atomic
                            object. defaults to False if not provided.

        :rtype: bool
        """

        if session_services.is_request_context_available() is True:
            has_atomic = self.atomic_registry.has()
            if atomic is True:
                return has_atomic
            return has_atomic or super().has()

        return False

    def set(self, obj, atomic=False):
        """
        sets the value for the current scope.

        :param object obj: object to be set in registry.

        :param bool atomic: specifies that it must set an atomic object.
                            defaults to False if not provided.
        """

        if atomic is True:
            self.atomic_registry.set(obj)
        else:
            super().set(obj)

    def clear(self, atomic=False):
        """
        clears the current scope's object, if any.

        it also clears if any atomic object is available.
        if `atomic=True` is provided, it only clears the
        current atomic object if available.

        :param bool atomic: specifies that it must only clear the current atomic
                            object of current scope. otherwise, it clears all objects
                            of current scope. defaults to False if not provided.
        """

        self.atomic_registry.clear(atomic)
        if atomic is False:
            super().clear()

    def clear_all_atomic(self):
        """
        clears all atomic objects of current scope, if any.
        """

        self.atomic_registry.clear()

    def get(self, atomic=False):
        """
        gets the current object of current scope if available, otherwise returns None.

        :param bool atomic: specifies that it must get the current atomic object of
                            current scope, otherwise it returns the normal object.
                            defaults to False if not provided.

        :rtype: object
        """

        if atomic is True:
            return self._get_atomic()

        return self._get()

    def get_all_atomic(self):
        """
        gets all atomic objects of current scope if available, otherwise returns an empty list.

        :rtype: list[object]
        """

        return self.atomic_registry.get_all()

    def _get_atomic(self):
        """
        gets the current atomic object of current scope if available, otherwise returns None.

        :rtype: object
        """

        return self.atomic_registry.get()

    def _get(self):
        """
        gets the current normal object of current scope if available, otherwise returns None.

        :rtype: object
        """

        try:
            return self.registry[self.scopefunc()]
        except KeyError:
            return None



# -*- coding: utf-8 -*-
"""
sentry exceptions module.
"""

from pyrin.core.exceptions import CoreException


class SentryManagerException(CoreException):
    """
    sentry manager exception.
    """
    pass


class SentryIntegrationNotFoundError(SentryManagerException):
    """
    sentry integration not found error.
    """
    pass


class InvalidSentryIntegrationTypeError(SentryManagerException):
    """
    invalid sentry integration type error.
    """
    pass


class DuplicatedSentryIntegrationError(SentryManagerException):
    """
    duplicated sentry integration error.
    """
    pass



# -*- coding: utf-8 -*-
"""
hashing handlers base module.
"""

import re

from abc import abstractmethod

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security.hashing.interface import AbstractHashingBase
from pyrin.settings.static import APPLICATION_ENCODING
from pyrin.utils import encoding
from pyrin.security.hashing.handlers.exceptions import InvalidHashedValueError, \
    HashingHandlerMismatchError, HashingHandlerException


class HashingBase(AbstractHashingBase):
    """
    hashing base class.
    all application hashing handlers must be subclassed from this.
    """

    # regular expression to validate format of full hashed values.
    # the following format will be matched: `$handler_name$other_parts$hashed_value`
    # note that different handlers may override this format in
    # their respective implementation.
    FORMAT_REGEX = re.compile(r'^\$[^$]+\$(.+)\$(.+)$')

    def __init__(self, **options):
        """
        initializes an instance of HashingBase.
        """

        super().__init__()

        # we set the algorithm of hashing handler as the name of it.
        self._set_name(self._get_algorithm(**options))
        self._encoding = APPLICATION_ENCODING

    def generate_hash(self, text, **options):
        """
        gets the hash of input text using a random or specified salt.

        :param str text: text to be hashed.

        :keyword bytes salt: salt to be used for hashing.
                             if not provided, a random salt will be generated
                             considering `salt_length` option.

        :keyword int rounds: rounds to perform for generating hash.
                             if not provided, default value from
                             relevant config will be used.

        :rtype: str
        """

        full_hashed_value = self._generate_hash(text, **options)
        return self._prepare_output(full_hashed_value, **options)

    @abstractmethod
    def _generate_hash(self, text, **options):
        """
        gets the hash of input text using a random or specified salt.

        :param str text: text to be hashed.

        :keyword bytes salt: salt to be used for hashing.
                             if not provided, a random salt will be generated
                             considering `salt_length` option.

        :keyword int rounds: rounds to perform for generating hash.
                             if not provided, default value from
                             relevant config will be used.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bytes
        """

        raise CoreNotImplementedError()

    def is_match(self, text, full_hashed_value, **options):
        """
        gets a value indicating that given text's
        hash is identical to given full hashed value.

        :param str text: text to be hashed.
        :param str full_hashed_value: full hashed value to compare with.

        :rtype: bool
        """

        try:
            self._validate_format(full_hashed_value, **options)
            hashed_part = self._get_hashed_part(self._prepare_input(full_hashed_value),
                                                **options)

            return self._is_match(text, hashed_part, **options)

        except HashingHandlerException:
            return False

    @abstractmethod
    def _is_match(self, text, hashed_value, **options):
        """
        gets a value indicating that given text's
        hash is identical to given hashed value.

        :param str text: text to be hashed.
        :param bytes hashed_value: hashed value to compare with.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_algorithm(self, **options):
        """
        gets the hashing algorithm.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def _get_separator(self):
        """
        gets the separator used between each part of this handler's hashed result.

        :rtype: bytes
        """

        return b'$'

    def _get_separator_count(self):
        """
        gets the separator count used between parts of this handler's hashed result.

        :rtype: int
        """

        return 3

    @abstractmethod
    def _extract_parts_from_final_hash(self, full_hashed_value, **options):
        """
        extracts different parts of given full hashed value.

        :param bytes full_hashed_value: full hashed value to extract it's parts.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: tuple
        """

        raise CoreNotImplementedError()

    def _validate_format(self, full_hashed_value, **options):
        """
        validates the format of full hashed value.
        an exception will be raised on invalid values.

        :param str full_hashed_value: full hashed value to be validated.

        :raises InvalidHashedValueError: invalid hashed value error.
        """

        if not self.FORMAT_REGEX.match(full_hashed_value):
            raise InvalidHashedValueError('Input value is not a valid '
                                          '[{current}] hashing value.'
                                          .format(current=self._get_algorithm()))

    @abstractmethod
    def _get_hashed_part(self, full_hashed_value, **options):
        """
        gets the hashed part from full hashed value which current handler understands it.

        :param bytes full_hashed_value: full hashed value to get hashed part from it.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bytes
        """

        raise CoreNotImplementedError()

    def _validate_handler(self, handler_name, **options):
        """
        validates the handler name to match the current handler.
        it will raise an error on handler mismatch.

        :param str handler_name: handler name to be validated.

        :raises HashingHandlerMismatchError: hashing handler mismatch error.
        """

        if handler_name != self._get_algorithm():
            raise HashingHandlerMismatchError('Hashing handler [{handler}] does not '
                                              'match the current handler which is [{current}].'
                                              .format(handler=handler_name,
                                                      current=self._get_algorithm()))

    def _prepare_output(self, full_hashed_value, **options):
        """
        prepares output value before returning from this handler.

        :param bytes full_hashed_value: full hashed value to be prepared.

        :rtype: str
        """

        return full_hashed_value.decode(self._encoding)

    def _prepare_input(self, full_hashed_value, **options):
        """
        prepares input value to be handled by this handler.

        :param str full_hashed_value: full hashed value to be prepared.

        :rtype: bytes
        """

        return full_hashed_value.encode(self._encoding)

    def _decode_hash_part(self, value):
        """
        decodes base64 encoded hash part into raw bytes.

        :param bytes value: base64 encoded value.

        :returns: raw bytes.
        :rtype: bytes
        """

        return encoding.base64_to_bytes(value)

    def _encode_hash_part(self, value):
        """
        encodes raw bytes hash part into base64 encoded bytes.

        :param bytes value: raw bytes value.

        :returns: base64 encoded bytes.
        :rtype: bytes
        """

        return encoding.bytes_to_base64(value)



# -*- coding: utf-8 -*-
"""
orm sql schema package.
"""

from pyrin.packaging.base import Package


class ORMSQLSchemaPackage(Package):
    """
    orm sql schema package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
template interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class TemplateHandlerSingletonMeta(MultiSingletonMeta):
    """
    template handler singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractTemplateHandler(CoreObject, metaclass=TemplateHandlerSingletonMeta):
    """
    abstract template handler class.
    """

    @abstractmethod
    def create(self, *args, **kwargs):
        """
        creates the template files in target path of this handler.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def name(self):
        """
        gets the name of this template handler.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
router handlers exceptions module.
"""

from pyrin.core.enumerations import ClientErrorResponseCodeEnum
from pyrin.core.exceptions import CoreException, CoreBusinessException
from pyrin.processor.request.wrappers.exceptions import BadRequestError
from pyrin.security.authentication.exceptions import AuthenticationFailedError


class RouterHandlerException(CoreException):
    """
    router handler exception.
    """
    pass


class RouterHandlerBusinessException(CoreBusinessException,
                                     RouterHandlerException):
    """
    router handler business exception.
    """
    pass


class InvalidViewFunctionTypeError(RouterHandlerException):
    """
    invalid view function type error.
    """
    pass


class MaxContentLengthLimitMismatchError(RouterHandlerException):
    """
    max content length limit mismatch error.
    """
    pass


class FreshAuthenticationRequiredError(AuthenticationFailedError,
                                       RouterHandlerBusinessException):
    """
    fresh authentication required error.
    """
    pass


class PermissionTypeError(RouterHandlerException):
    """
    permission type error.
    """
    pass


class InvalidResultSchemaTypeError(RouterHandlerException):
    """
    invalid result schema type error.
    """
    pass


class RouteIsNotBoundedError(RouterHandlerException):
    """
    route is not bounded error.
    """
    pass


class RouteIsNotBoundedToMapError(RouterHandlerException):
    """
    route is not bounded to map error.
    """
    pass


class InvalidResponseStatusCodeError(RouterHandlerException):
    """
    invalid response status code error.
    """
    pass


class ViewFunctionParamsError(BadRequestError,
                              RouterHandlerBusinessException):
    """
    view function params error.
    """
    pass


class ViewFunctionRequiredParamsError(ViewFunctionParamsError):
    """
    view function required params error.
    """
    pass


class InvalidRequestLimitError(RouterHandlerException):
    """
    invalid request limit error.
    """
    pass


class InvalidLifetimeError(RouterHandlerException):
    """
    invalid lifetime error.
    """
    pass


class RequestLimitOrLifetimeRequiredError(RouterHandlerException):
    """
    request limit or lifetime required error.
    """
    pass


class URLNotFoundError(RouterHandlerBusinessException):
    """
    url not found error.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of URLNotFoundError.

        :keyword dict data: extra data for exception.
        """

        super().__init__(*args, **kwargs)
        self._code = ClientErrorResponseCodeEnum.NOT_FOUND



# -*- coding: utf-8 -*-
"""
utils path module.
"""

import os
import sys
import shutil

from pyrin.utils.exceptions import PathIsNotAbsoluteError, InvalidPathError, \
    PathNotExistedError, PathAlreadyExistedError


def get_module_file_path(module_name):
    """
    gets the absolute file path of module with given name.

    :param str module_name: module name to get its file path.

    :rtype: str
    """

    return os.path.abspath(sys.modules[module_name].__file__)


def get_main_package_name(module_name):
    """
    gets the main package name from given module name.

    for example for `pyrin.database.manager` module, it
    returns `pyrin` as the main package name.

    :param str module_name: module name to get its root package name.

    :rtype: str
    """

    return module_name.split('.')[0]


def get_main_package_path(module_name):
    """
    gets the absolute path of the main package of module with given name.

    :param str module_name: module name to get its main package path.

    :rtype: str
    """

    relative_module_path = module_name.replace('.', os.path.sep)
    root_package = get_main_package_name(module_name)
    absolute_module_path = get_module_file_path(module_name)
    temp_absolute_module_path = absolute_module_path.replace(relative_module_path, '*')
    excess_part = temp_absolute_module_path.split('*')[-1]
    list_path = list(temp_absolute_module_path)

    for i in range(-1, -len(list_path), -1):
        if list_path[i] == '*':
            list_path[i] = root_package
            break

    main_package_path = ''.join(list_path)
    main_package_path = main_package_path.replace('*', relative_module_path)
    main_package_path = main_package_path.replace(excess_part, '').rstrip(os.path.sep)

    return main_package_path


def get_pyrin_main_package_name():
    """
    gets the name of pyrin main package name.

    it would always be `pyrin` in normal cases.

    :rtype: str
    """

    return get_main_package_name(__name__)


def get_pyrin_main_package_path():
    """
    gets the absolute path of pyrin main package.

    :rtype: str
    """

    return get_main_package_path(__name__)


def get_package_name(path, root_path):
    """
    gets the full package name for provided path.

    :param str path: full path of package.
                     example path = `/home/src/pyrin/database`.

    :param str root_path: root path in which this path is located.
                          example root_path = `/home/src`

    :rtype: str
    """

    return path.replace(root_path, '').replace(os.path.sep, '.').lstrip('.')


def get_package_path(module_name):
    """
    gets the absolute path of the package of module with given name.

    :param str module_name: module name to get its package path.

    :rtype: str
    """

    absolute_module_path = get_module_file_path(module_name)
    parts = os.path.split(absolute_module_path)
    return parts[0]


def create_directory(target, ignore_existed=False):
    """
    creates a directory with given absolute target path.

    it will create all not available intermediate directories as well.

    :param str target: absolute path of directory to be created.

    :param bool ignore_existed: specifies that if the directory is
                                already existed, ignore it. otherwise
                                raise an error. defaults to False

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathAlreadyExistedError: path already existed error.
    """

    if ignore_existed is not True:
        assert_not_exists(target)

    os.makedirs(target, exist_ok=ignore_existed)


def copy_file(source, target):
    """
    copies the given source file into given target file or directory.

    note that target could also be a directory, if so,
    then the file with the source name will be generated.
    both source and target paths must be absolute.

    :param str source: source file absolute path.
    :param str target: target file or directory absolute path.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    """

    assert_exists(source)
    assert_absolute(target)
    shutil.copy2(source, target)


def remove_file(source):
    """
    removes the given source file.

    :param str source: source file absolute path.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    """

    assert_exists(source)
    os.remove(source)


def copy_directory(source, target, ignore_existed=False, ignore=None):
    """
    copies the given source directory contents into given target directory.

    both source and target paths must be absolute.

    :param str source: source directory absolute path.
    :param str target: target directory absolute path.

    :param bool ignore_existed: specifies that if the destination directory
                                is already existed, it should not raise an error.
                                defaults to False if not provided.

    :param callable ignore: the optional ignore argument is a callable. if given, it
                            is called with the `source` parameter, which is the directory
                            being visited by copytree(), and `names` which is the list of
                            `source` contents, as returned by os.listdir():
                            callable(src, names) -> ignored_names

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    """

    assert_exists(source)
    assert_absolute(target)
    copytree_ex(source, target, ignore=ignore, ignore_existed=ignore_existed)


def assert_absolute(source):
    """
    asserts that given source path is absolute.

    :param str source: source path to be checked.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    """

    if source is None:
        raise InvalidPathError('Provided path could not be None.')

    if not os.path.isabs(source):
        raise PathIsNotAbsoluteError('Provided path [{source}] must be absolute.'
                                     .format(source=source))


def exists(source):
    """
    gets a value indicating that given source path exists on file system.

    :param str source: source path to be checked for existence.
                       it must be an absolute path.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    """

    assert_absolute(source)
    return os.path.exists(source)


def assert_exists(source):
    """
    asserts that given source path exists on file system.

    :param str source: source path to be checked for existence.
                       it must be an absolute path.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathNotExistedError: path not existed error.
    """

    if not exists(source):
        raise PathNotExistedError('Provided path [{source}] does not exist.'
                                  .format(source=source))


def assert_not_exists(source):
    """
    asserts that given source path not exists on file system.

    :param str source: source path to be checked for not existence.
                       it must be an absolute path.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.
    :raises PathAlreadyExistedError: path already existed error.
    """

    if exists(source):
        raise PathAlreadyExistedError('Provided path [{source}] already existed.'
                                      .format(source=source))


def get_first_available_file(*paths, file_name):
    """
    gets the first path which a file with given name is resided in it.

    it returns None if the file is not available in any of given paths.

    :param str paths: paths to look for file in them.
                      all paths must be absolute.

    :param str file_name: file name with extension to look for.

    :raises InvalidPathError: invalid path error.
    :raises PathIsNotAbsoluteError: path is not absolute error.

    :rtype: str
    """

    for single_path in paths:
        assert_absolute(single_path)
        file_path = os.path.abspath(os.path.join(single_path, file_name))
        if os.path.isfile(file_path):
            return file_path

    return None


def copytree_ex(source, destination, symlinks=False, ignore=None,
                copy_function=shutil.copy2, ignore_dangling_symlinks=False,
                ignore_existed=False):
    """
    recursively copy a directory tree.

    this method patches the `shutil.copytree()` method to let the caller decide
    to let or not to let to copy if destination directory is already existed.
    if exception(s) occur, an error is raised with a list of reasons.

    :param str source: source directory.
    :param str destination: destination directory.

    :param bool symlinks: if the optional symlinks flag is True, symbolic links in the
                          source tree result in symbolic links in the destination tree.
                          if it is false, the contents of the files pointed to by symbolic
                          links are copied. if the file pointed by the symlink doesn't
                          exist, an exception will be added in the list of errors raised in
                          an error exception at the end of the copy process. defaults to
                          False if not provided.

    :param callable ignore: the optional ignore argument is a callable. if given, it
                            is called with the `source` parameter, which is the directory
                            being visited by copytree(), and `names` which is the list of
                            `source` contents, as returned by os.listdir():
                            callable(src, names) -> ignored_names

    :param callable copy_function: is a callable that will be used to copy each file.
                                   it will be called with the source path and the
                                   destination path as arguments. by default, `copy2()`
                                   is used, but any function that supports the same
                                   signature can be used.

    :param bool ignore_dangling_symlinks: if set to True, the exception for broken
                                          symlinks will not be raised.
                                          defaults to False if not provided.

    :param bool ignore_existed: specifies that if the destination directory
                                is already existed, it should not raise an error.
                                defaults to False if not provided.
    """

    names = os.listdir(source)
    if ignore is not None:
        ignored_names = ignore(source, names)
    else:
        ignored_names = set()

    os.makedirs(destination, exist_ok=ignore_existed)
    errors = []
    for name in names:
        if name in ignored_names:
            continue
        source_name = os.path.join(source, name)
        real_name = name
        if real_name.endswith('-py'):
            real_name = real_name.replace('-py', '.py')
        destination_name = os.path.join(destination, real_name)
        try:
            if os.path.islink(source_name):
                link_to = os.readlink(source_name)
                if symlinks:
                    # We can't just leave it to `copy_function` because legacy
                    # code with a custom `copy_function` may rely on copytree
                    # doing the right thing.
                    os.symlink(link_to, destination_name)
                    shutil.copystat(source_name, destination_name,
                                    follow_symlinks=not symlinks)
                else:
                    # ignore dangling symlink if the flag is on
                    if not os.path.exists(link_to) and ignore_dangling_symlinks:
                        continue
                    # otherwise let the copy occurs. copy2 will raise an error
                    if os.path.isdir(source_name):
                        copytree_ex(source_name, destination_name, symlinks, ignore,
                                    copy_function)
                    else:
                        copy_function(source_name, destination_name)
            elif os.path.isdir(source_name):
                copytree_ex(source_name, destination_name, symlinks, ignore, copy_function)
            else:
                # Will raise a SpecialFileError for unsupported file types
                copy_function(source_name, destination_name)
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except shutil.Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((source_name, destination_name, str(why)))
    try:
        shutil.copystat(source, destination)
    except OSError as why:
        # Copying file access times may fail on Windows
        if getattr(why, 'winerror', None) is None:
            errors.append((source, destination, str(why)))
    if errors:
        raise shutil.Error(errors)


def get_pycache(source, names):
    """
    gets all `__pycache__` directories available in names.

    :param str source: source directory of contents.
    :param list[str] names: name of all contents in source.

    :rtype: list[str]
    """

    return [name for name in names if '__pycache__' in name]



# -*- coding: utf-8 -*-
"""
authentication component module.
"""

from pyrin.application.decorators import component
from pyrin.security.authentication import AuthenticationPackage
from pyrin.security.authentication.manager import AuthenticationManager
from pyrin.application.structs import Component


@component(AuthenticationPackage.COMPONENT_NAME)
class AuthenticationComponent(Component, AuthenticationManager):
    """
    authentication component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
alembic handlers merge module.
"""

from pyrin.database.migration.alembic.enumerations import AlembicCLIHandlersEnum
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase
from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.handlers.params import MessageParam, \
    BranchLabelParam, RevisionIDParam, RevisionsParam


@alembic_cli_handler()
class MergeCLIHandler(AlembicCLIHandlerBase):
    """
    merge cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of MergeCLIHandler.
        """

        super().__init__(AlembicCLIHandlersEnum.MERGE)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([MessageParam(), BranchLabelParam(),
                       RevisionIDParam(), RevisionsParam()])

        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
caching structs module.
"""

from collections import OrderedDict

from pyrin.core.globals import LIST_TYPES
from pyrin.caching.globals import NOT_HASHABLE_TYPES


class CacheableDict(OrderedDict):
    """
    cacheable dict class.

    this class is a normal ordered dict but it is hashable and could
    be used as dict key.
    """

    def __init__(self, mapping=None, **kwargs):
        """
        initializes an instance of CacheableDict.

        :param dict | CacheableDict | iterable mapping: values to create a dict form them.
        """

        cacheable_kwargs = {}
        if len(kwargs) > 0:
            cacheable_kwargs = self._convert_dict(kwargs)

        if isinstance(mapping, CacheableDict):
            super().__init__(self._sort_list(self._union_lists(mapping.items(),
                                                               cacheable_kwargs.items())))
        elif isinstance(mapping, dict):
            temp = {}
            for key, value in mapping.items():
                if isinstance(value, NOT_HASHABLE_TYPES):
                    temp[key] = self._convert(value)
                else:
                    temp[key] = value

            super().__init__(self._sort_list(self._union_lists(temp.items(),
                                                               cacheable_kwargs.items())))

        elif isinstance(mapping, LIST_TYPES):
            temp = []
            for key, value in mapping:
                if isinstance(value, NOT_HASHABLE_TYPES):
                    temp.append((key, self._convert(value)))
                else:
                    temp.append((key, value))

            super().__init__(self._sort_list(self._union_lists(temp,
                                                               cacheable_kwargs.items())))

        else:
            super().__init__(self._sort_list(cacheable_kwargs.items()))

    def __hash__(self):
        """
        gets the hash of current dict.

        :rtype: int
        """

        return hash(tuple(self.items()))

    def __eq__(self, other):
        """
        implements the equality operator.

        :param object other: other instance to check for equality.

        :rtype: bool
        """

        if not isinstance(other, CacheableDict):
            return False

        self_len = len(self)
        other_len = len(other)
        if self_len != other_len:
            return False

        if self_len == 0:
            return True

        return hash(self) == hash(other)

    def __ne__(self, other):
        """
        implements the not equality operator.

        :param object other: other instance to check for not equality.

        :rtype: bool
        """

        return not self == other

    def _union_lists(self, first, second):
        """
        gets the union of two iterables.

        :param iterable first: first iterable.
        :param iterable second: second iterable.

        :rtype: list
        """

        first = list(first)
        second = list(second)
        first.extend(second)
        return first

    def _convert(self, value):
        """
        converts all required values of given object to be hashable.

        if given object is not a dict or collection it returns the same input value.

        :param dict | list | tuple | set value: a dict or collection.

        :rtype: CacheableDict | tuple | object
        """

        if isinstance(value, dict):
            return self._convert_dict(value)
        elif isinstance(value, LIST_TYPES):
            return self._convert_list(value)
        return value

    def _convert_dict(self, value):
        """
        converts all required values of given dict to be hashable.

        :param dict value: dict to convert its value.

        :rtype: CacheableDict
        """

        return CacheableDict(value)

    def _convert_list(self, items):
        """
        converts all required values of different items of given iterable.

        :param list | tuple | set items: items to convert their values.

        :rtype: tuple
        """

        results = []
        for item in items:
            results.append(self._convert(item))

        return tuple(self._sort_list(results))

    def _sort_list(self, value):
        """
        returns the sorted list.

        if values are not comparable, it returns an unsorted list.

        :param iterable value: value ot be sorted.

        :rtype: list
        """

        try:
            # try to sort by value if all values are same type.
            return sorted(value)
        except TypeError:
            # fallback to unsorted values when value types are incomparable.
            return list(value)



# -*- coding: utf-8 -*-
"""
mimetype enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class MIMETypeEnum(CoreEnum):
    """
    mimetype enum.
    """

    TEXT = 'text/plain'
    HTML = 'text/html'
    JAVASCRIPT = 'text/javascript'
    OCTET_STREAM = 'application/octet-stream'
    JSON = 'application/json'



# -*- coding: utf-8 -*-
"""
locale manager module.
"""

from flask_babel import Babel
from babel import localedata

import pyrin.security.session.services as session_services
import pyrin.configuration.services as config_services
import pyrin.globalization.datetime.services as datetime_services

from pyrin.core.structs import Manager
from pyrin.application.services import get_current_app
from pyrin.globalization.locale import LocalePackage
from pyrin.globalization.locale.exceptions import InvalidLocaleSelectorTypeError, \
    InvalidTimezoneSelectorTypeError, LocaleSelectorHasBeenAlreadySetError, \
    TimezoneSelectorHasBeenAlreadySetError


class LocaleManager(Manager):
    """
    locale manager class.
    """

    package_class = LocalePackage

    def __init__(self, **options):
        """
        initializes an instance of LocaleManager.
        """

        super().__init__()

        self._babel = Babel(get_current_app())
        self._default_locale = config_services.get('globalization', 'locale',
                                                   'babel_default_locale')

    def set_locale_selector(self, func):
        """
        sets the given function as locale selector.

        :param function func: function to be set as locale selector.

        :raises InvalidLocaleSelectorTypeError: invalid locale selector type error.

        :raises LocaleSelectorHasBeenAlreadySetError: locale selector has been
                                                      already set error.
        """

        if not callable(func):
            raise InvalidLocaleSelectorTypeError('Input parameter [{locale}] '
                                                 'is not callable.'
                                                 .format(locale=func))

        if self._babel.locale_selector_func is not None:
            raise LocaleSelectorHasBeenAlreadySetError('Locale selector has been already '
                                                       'set and could not be overwritten.')

        self._babel.locale_selector_func = func

    def set_timezone_selector(self, func):
        """
        sets the given function as timezone selector.

        :param function func: function to be set as timezone selector.

        :raises InvalidTimezoneSelectorTypeError: invalid timezone selector type error.

        :raises TimezoneSelectorHasBeenAlreadySetError: timezone selector has been
                                                        already set error.
        """

        if not callable(func):
            raise InvalidTimezoneSelectorTypeError('Input parameter [{timezone}] '
                                                   'is not callable.'
                                                   .format(timezone=func))

        if self._babel.timezone_selector_func is not None:
            raise TimezoneSelectorHasBeenAlreadySetError('Timezone selector has been already '
                                                         'set and could not be overwritten.')

        self._babel.timezone_selector_func = func

    def get_default_locale(self):
        """
        gets the default locale of application from `globalization` config store.

        :rtype: str
        """

        return self._default_locale

    def get_current_locale(self):
        """
        gets the current locale that should be used for current request.

        it never raises an error and returns the default locale if anything goes wrong.

        :rtype: str
        """

        request = session_services.get_safe_current_request()
        locale = None
        if request is not None:
            locale = request.locale

        return locale or self._default_locale

    def get_current_timezone(self):
        """
        gets the current timezone that should be used for current request.

        it never raises an error and returns the default timezone if anything goes wrong.

        :rtype: str
        """

        return datetime_services.get_timezone_name(server=False)

    def locale_exists(self, locale_name):
        """
        gets a value indicating that a locale with the given name exists.

        :param str locale_name: locale name to check for existence.

        :rtype: bool
        """

        return localedata.exists(locale_name)



# -*- coding: utf-8 -*-
"""
api services module.
"""

from werkzeug.exceptions import HTTPException

from pyrin.application.decorators import error_handler
from pyrin.application.services import get_component
from pyrin.core.exceptions import CoreException, CoreBusinessException
from pyrin.api import APIPackage


@error_handler(HTTPException)
def handle_http_error(exception):
    """
    handles http exceptions.

    note that normally you should never call this method manually.

    :param HTTPException exception: exception instance.

    :returns: tuple[dict | object, int, CoreHeaders]
    :rtype: tuple
    """

    return get_component(APIPackage.COMPONENT_NAME).handle_http_error(exception)


@error_handler(CoreBusinessException)
def handle_server_business_error(exception):
    """
    handles server internal core business exceptions.

    note that normally you should never call this method manually.

    :param CoreBusinessException exception: core business exception instance.

    :returns: tuple[dict | object, int, CoreHeaders]
    :rtype: tuple
    """

    return get_component(APIPackage.COMPONENT_NAME).handle_server_business_error(exception)


@error_handler(CoreException)
def handle_server_error(exception):
    """
    handles server internal core exceptions.

    note that normally you should never call this method manually.
    in any environment which debug mode is False, the original error
    message will be replaced by a generic error message before being
    sent to client for security reasons.

    :param CoreException exception: core exception instance.

    :returns: tuple[dict | object, int, CoreHeaders]
    :rtype: tuple
    """

    return get_component(APIPackage.COMPONENT_NAME).handle_server_error(exception)


@error_handler(Exception)
def handle_server_unknown_error(exception):
    """
    handles unknown server internal exceptions.

    note that normally you should never call this method manually.
    in any environment which debug mode is False, the original error
    message will be replaced by a generic error message before being
    sent to client for security reasons.

    :param Exception exception: exception instance.

    :returns: tuple[dict | object, int, CoreHeaders]
    :rtype: tuple
    """

    return get_component(APIPackage.COMPONENT_NAME).handle_server_unknown_error(exception)


def register_hook(instance):
    """
    registers the given instance into api hooks.

    :param APIHookBase instance: api hook instance to be registered.

    :raises InvalidAPIHookTypeError: invalid api hook type error.
    """

    get_component(APIPackage.COMPONENT_NAME).register_hook(instance)



# -*- coding: utf-8 -*-
"""
cli mixin param module.
"""

from copy import deepcopy

from pyrin.cli.params import CLIParamBase, HelpParam
from pyrin.core.globals import LIST_TYPES
from pyrin.cli.arguments import PositionalArgument
from pyrin.core.structs import CoreObject
from pyrin.cli.exceptions import PositionalArgumentsIndicesError, InvalidCLIParamTypeError


class CLIParamMixin(CoreObject):
    """
    cli param mixin class.

    every cli handler class that needs to manage
    cli parameters must be subclassed from this.
    """

    def __init__(self, **options):
        """
        initializes an instance of CLIParamMixin.

        :keyword bool help_param: specifies that help param must
                                  be added for this handler.
                                  defaults to True if not provided.

        :raises PositionalArgumentsIndicesError: positional arguments indices error.
        """

        super().__init__()

        self._params = []
        help_param = options.get('help_param', True)
        params = self.__get_params(help_param)
        self._process_params(params)

    def _process_params(self, params):
        """
        processes the cli params that should be added.

        :param list[CLIParamBase] params: cli params of this handler.

        :raises InvalidCLIParamTypeError: invalid cli param type error.
        :raises PositionalArgumentsIndicesError: positional arguments indices error.
        """

        self.__extend_params(params)
        self._validate_positional_arguments()

    def _validate_positional_arguments(self):
        """
        validates the positional arguments.

        it will make sure that their indices are correct.

        :raises PositionalArgumentsIndicesError: positional arguments indices error.
        """

        positionals = [item for item in self._params
                       if isinstance(item, PositionalArgument)
                       and item.validate_index is not False]

        length = len(positionals)
        if length > 0:
            indices = [item.index for item in positionals]
            required_indices = list(range(0, length))

            if sorted(indices) != sorted(required_indices):
                raise PositionalArgumentsIndicesError('CLI handler [{handler}] has {length} '
                                                      'positional arguments {args}, but '
                                                      'indices of positional arguments are '
                                                      'incorrect. indices must be in the '
                                                      'range of {indices} for each argument '
                                                      'and it must be unique per argument. '
                                                      'the current indices are {current}.'
                                                      .format(handler=self,
                                                              length=length,
                                                              args=positionals,
                                                              indices=required_indices,
                                                              current=indices))

    def __add_param(self, item):
        """
        adds the given item into params list.

        :param CLIParamBase item: cli param to be added.

        :raises InvalidCLIParamTypeError: invalid cli param type error.
        """

        if not isinstance(item, CLIParamBase):
            raise InvalidCLIParamTypeError('Input parameter [{param}] is not '
                                           'an instance of [{instance}].'
                                           .format(param=item,
                                                   instance=CLIParamBase))

        self._params.append(item)

    def __extend_params(self, items):
        """
        extends params list with the given items.

        :param list[CLIParamBase] items: cli params to be added.

        :raises InvalidCLIParamTypeError: invalid cli param type error.
        """

        for item in items:
            self.__add_param(item)

    def _process_inputs(self, **options):
        """
        processes the inputs in given dict with all available params.

        :rtype: dict
        """

        processed_inputs = deepcopy(options)
        for item in self._params:
            processed_inputs = item.process_inputs(**processed_inputs)

        return processed_inputs

    def _bind_cli_arguments(self, commands, **options):
        """
        binds cli arguments with values available in options.

        it adds items into the exact input list.
        options are a mapping of real python method inputs.

        :param list[object] commands: cli commands list.

        :returns: list of all commands in the form of:
                  [str name1, str value1, ...] or
                  [str name1, str value1, str value2, ...]

        :rtype: list
        """

        start = len(commands)
        for param in self._params:
            real_value = options.get(param.param_name, None)
            representation = param.get_representation(real_value)

            if representation is None:
                continue

            if isinstance(param, PositionalArgument):
                if isinstance(representation, LIST_TYPES):
                    self._inject_positional_arguments(commands, representation,
                                                      param.index + start)
                else:
                    commands.insert(param.index + start, representation)
            elif isinstance(representation, LIST_TYPES):
                commands.extend(representation)
            else:
                commands.append(representation)

    def _inject_positional_arguments(self, commands, arguments, index):
        """
        injects (in-place) given arguments into specified index of given commands list.

        :param list[object] commands: list of cli commands.

        :param list[object] arguments: arguments that must be
                                       injected into commands list.

        :param index: zero-based index in which arguments must be injected.
        """

        for item_index, item in enumerate(arguments):
            commands.insert(index + item_index, item)

    def __get_params(self, help_param=True):
        """
        gets the list of current handler's prams.

        :param bool help_param: specifies that help param
                                must also be added to params list.
                                defaults to True if not provided.

        :rtype: list[CLIParamBase]
        """

        params = []
        if help_param is True:
            params.append(HelpParam())

        self._inject_params(params)
        return params

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        all subclasses that want to override this method to inject
        their params into the given list must call `super()._inject_params(params)`
        after their work has been done.

        :param list[CLIParamBase] params: list of all params.
        """
        pass



# -*- coding: utf-8 -*-
"""
validator auto manager module.
"""

from uuid import UUID
from decimal import Decimal
from datetime import datetime, date, time

import pyrin.validator.services as validator_services
import pyrin.database.model.services as model_services

from pyrin.core.structs import Manager, Context
from pyrin.validator.auto import ValidatorAutoPackage
from pyrin.validator.auto.handler import AutoValidator
from pyrin.validator.handlers.base import ValidatorBase
from pyrin.validator.handlers.dictionary import DictionaryValidator
from pyrin.validator.handlers.uuid import UUIDValidator
from pyrin.validator.handlers.string import StringValidator
from pyrin.validator.handlers.boolean import BooleanValidator
from pyrin.validator.handlers.datetime import DateTimeValidator, DateValidator, TimeValidator
from pyrin.validator.handlers.number import IntegerValidator, FloatValidator, DecimalValidator
from pyrin.validator.handlers.misc import MaximumValidator, MinimumValidator, \
    RangeValidator, InValidator, NotInValidator


class ValidatorAutoManager(Manager):
    """
    validator auto manager class.
    """

    package_class = ValidatorAutoPackage

    def __init__(self):
        """
        initializes an instance of ValidatorAutoManager.
        """

        super().__init__()

        # a dictionary containing python type to validator map.
        # example: {type | str python_type_or_name: type[AbstractValidatorBase] validator_class}
        self._type_to_validator_map = Context()
        self._initialize_map()

    def _initialize_map(self):
        """
        initializes the python type to validator map.
        """

        self._type_to_validator_map[int] = IntegerValidator
        self._type_to_validator_map[float] = FloatValidator
        self._type_to_validator_map[Decimal] = DecimalValidator
        self._type_to_validator_map[bool] = BooleanValidator
        self._type_to_validator_map[str] = StringValidator
        self._type_to_validator_map[dict] = DictionaryValidator
        self._type_to_validator_map[datetime] = DateTimeValidator
        self._type_to_validator_map[date] = DateValidator
        self._type_to_validator_map[time] = TimeValidator
        self._type_to_validator_map[UUID] = UUIDValidator
        self._type_to_validator_map['max'] = MaximumValidator
        self._type_to_validator_map['min'] = MinimumValidator
        self._type_to_validator_map['range'] = RangeValidator
        self._type_to_validator_map['in'] = InValidator
        self._type_to_validator_map['not_in'] = NotInValidator

    def _get_validator_class(self, python_type):
        """
        gets the validator class for given python type.

        it return None if no validator found with given type or name.

        :param type | str python_type: python type or name of validator class.

        :rtype: type[AbstractValidatorBase]
        """

        return self._type_to_validator_map.get(python_type)

    def _get_type_validator(self, domain, field, **options):
        """
        gets the required type validator for given field.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.

        :rtype: AbstractValidatorBase
        """

        collection_type, python_type = field.get_python_type()
        if collection_type is list:
            options.update(is_list=True)

        validator_class = self._get_validator_class(python_type)
        if validator_class is not None:
            return validator_class(domain, field, **options)

        return ValidatorBase(domain, field, **options)

    def _get_range_validator(self, domain, field, **options):
        """
        gets the required range validator for given field.

        it either returns a min validator, max validator, range validator or None.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.

        :rtype: AbstractValidatorBase
        """

        validator_class = None
        if field.min_value is not None and field.max_value is not None:
            validator_class = self._get_validator_class('range')

        elif field.min_value is not None:
            validator_class = self._get_validator_class('min')

        elif field.max_value is not None:
            validator_class = self._get_validator_class('max')

        if validator_class is not None:
            return validator_class(domain, field, **options)

        return None

    def _get_in_validator(self, domain, field, **options):
        """
        gets the required in validator for given field.

        it either returns an in validator, not in validator or None.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.

        :rtype: AbstractValidatorBase
        """

        validator_class = None
        if field.check_in is not None:
            validator_class = self._get_validator_class('in')

        elif field.check_not_in is not None:
            validator_class = self._get_validator_class('not_in')

        if validator_class is not None:
            return validator_class(domain, field, **options)

        return None

    def register_auto_validator(self, domain, field, **options):
        """
        register required auto validator for given field.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.
        """

        type_validator = self._get_type_validator(domain, field, **options)
        range_validator = self._get_range_validator(domain, field, **options)
        in_validator = self._get_in_validator(domain, field, **options)
        custom_validator = validator_services.try_get_validator(domain, field.key)
        if custom_validator is not None:
            options.update(replace=True)

        options.update(type_validator=type_validator,
                       range_validator=range_validator,
                       in_validator=in_validator,
                       custom_validator=custom_validator)

        auto_validator = AutoValidator(domain, field, **options)
        validator_services.register_validator(auto_validator, **options)

    def register_auto_validators(self):
        """
        registers all auto validators.

        it returns the count of registered auto validators.

        :rtype: int
        """

        entities = model_services.get_entities()
        registered = 0
        for entity in entities:
            for column in entity.all_instrumented_attributes:
                if column.validated is True:
                    self.register_auto_validator(entity, column)
                    registered += 1

        return registered



# -*- coding: utf-8 -*-
"""
babel handlers init module.
"""

from pyrin.globalization.locale.babel.enumerations import BabelCLIHandlersEnum
from pyrin.globalization.locale.babel.interface import BabelCLIHandlerBase
from pyrin.globalization.locale.babel.decorators import babel_cli_handler
from pyrin.globalization.locale.babel.handlers.params import DomainParam, \
    InputTemplateFileParam, OutputDirectoryParam, LocaleParam, WidthParam, NoWrapParam, \
    OutputFileParam


@babel_cli_handler()
class InitCLIHandler(BabelCLIHandlerBase):
    """
    init cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of InitCLIHandler.
        """

        super().__init__(BabelCLIHandlersEnum.INIT)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([DomainParam(), InputTemplateFileParam(),
                       OutputDirectoryParam(), LocaleParam(),
                       WidthParam(), NoWrapParam(), OutputFileParam()])

        return super()._inject_params(params)
